"""Module with repository implementation for SQLAlchemy"""

import copy
import json
import logging
import uuid
from enum import Enum
from typing import Any

import sqlalchemy.dialects.postgresql as psql
from sqlalchemy import Column, MetaData, and_, create_engine, or_, orm, text
from sqlalchemy import types as sa_types
from sqlalchemy.engine.url import make_url
from sqlalchemy.exc import DatabaseError
from sqlalchemy.types import CHAR, TypeDecorator

from protean.container import Options
from protean.core.model import BaseModel
from protean.core.queryset import ResultSet
from protean.core.value_object import BaseValueObject
from protean.exceptions import (
    ConfigurationError,
    ObjectNotFoundError,
)
from protean.fields import (
    Auto,
    Boolean,
    Date,
    DateTime,
    Dict,
    Field,
    Float,
    Identifier,
    Integer,
    List,
    String,
    Text,
    ValueObject,
)
from protean.fields.association import Reference, _ReferenceField
from protean.fields.embedded import _ShadowField
from protean.globals import current_domain, current_uow
from protean.port.dao import BaseDAO, BaseLookup
from protean.port.provider import BaseProvider
from protean.reflection import attributes, id_field
from protean.utils import IdentityType
from protean.utils.query import Q

logging.getLogger("sqlalchemy").setLevel(logging.ERROR)
logger = logging.getLogger(__name__)


class GUID(TypeDecorator):
    """Platform-independent GUID type.

    Uses PostgreSQL's UUID type, otherwise uses
    CHAR(32), storing as stringified hex values.

    """

    impl = CHAR
    cache_ok = True

    def load_dialect_impl(self, dialect):
        if dialect.name == "postgresql":
            return dialect.type_descriptor(psql.UUID())
        else:
            return dialect.type_descriptor(CHAR(32))

    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        elif dialect.name == "postgresql":
            return str(value)
        else:
            if not isinstance(value, uuid.UUID):
                return "%.32x" % uuid.UUID(value).int
            else:
                # hexstring
                return "%.32x" % value.int

    def process_result_value(self, value, dialect):
        if value is None:
            return value
        else:
            if not isinstance(value, uuid.UUID):
                value = uuid.UUID(value)
            return value


def _get_identity_type():
    """Retrieve the configured data type for AutoGenerated Identifiers

    If `current_domain` is not yet available, it simply means that Protean is still being loaded.
    Default to `Identity.STRING`
    """
    if current_domain.config["identity_type"] == IdentityType.INTEGER.value:
        return sa_types.Integer
    elif current_domain.config["identity_type"] == IdentityType.STRING.value:
        return sa_types.String
    elif current_domain.config["identity_type"] == IdentityType.UUID.value:
        return GUID
    else:
        raise ConfigurationError(
            f'Unknown Identity Type {current_domain.config["identity_type"]}'
        )


def _default(value):
    """A function that gets called for objects that can’t otherwise be serialized.
    We handle the special case of Value Objects here.

    `TypeError` is raised for unknown types.
    """
    if isinstance(value, BaseValueObject):
        return value.to_dict()
    raise TypeError()


def _custom_json_dumps(value):
    """Custom JSON Serializer method to handle the special case of ValueObject deserialization.

    This method is passed into sqlalchemy as a value for param `json_serializer` in the call to `create_engine`.
    """
    return json.dumps(value, default=_default)


def derive_schema_name(model_cls):
    # Retain schema name if already present, otherwise derive from entity class
    if (
        hasattr(model_cls.meta_, "schema_name")
        and model_cls.meta_.schema_name is not None
    ):
        return model_cls.meta_.schema_name
    else:
        return model_cls.meta_.entity_cls.meta_.schema_name


class SqlalchemyModel(orm.DeclarativeBase, BaseModel):
    """Model representation for the Sqlalchemy Database"""

    def __init_subclass__(subclass, **kwargs):  # noqa: C901
        field_mapping = {
            Boolean: sa_types.Boolean,
            Date: sa_types.Date,
            DateTime: sa_types.DateTime,
            Dict: sa_types.PickleType,
            Float: sa_types.Float,
            Identifier: _get_identity_type(),
            Integer: sa_types.Integer,
            List: sa_types.PickleType,
            String: sa_types.String,
            Text: sa_types.Text,
            _ReferenceField: _get_identity_type(),
            ValueObject: sa_types.PickleType,
        }

        def field_mapping_for(field_obj: Field):
            """Return SQLAlchemy-equivalent type for Protean's field"""
            field_cls = type(field_obj)

            if field_cls is Auto:
                if field_obj.increment is True:
                    return sa_types.Integer
                else:
                    return _get_identity_type()

            return field_mapping.get(field_cls)

        # Update the class attrs with the entity attributes
        if "meta_" in subclass.__dict__:
            entity_cls = subclass.__dict__["meta_"].entity_cls
            for _, field_obj in attributes(entity_cls).items():
                attribute_name = field_obj.attribute_name

                # Map the field if not in attributes
                if attribute_name not in subclass.__dict__:
                    # Derive field based on field enclosed within ShadowField
                    if isinstance(field_obj, _ShadowField):
                        field_obj = field_obj.field_obj

                    field_cls = type(field_obj)
                    type_args = []
                    type_kwargs = {}

                    # Get the SA type
                    sa_type_cls = field_mapping_for(field_obj)

                    # Upgrade to Postgresql specific Data Types
                    if subclass.__dict__["engine"].dialect.name == "postgresql":
                        if field_cls == Dict and not field_obj.pickled:
                            sa_type_cls = psql.JSON

                        if field_cls == List and not field_obj.pickled:
                            sa_type_cls = psql.ARRAY

                            # Associate Content Type
                            if field_obj.content_type:
                                # Treat `ValueObject` differently because it is a field object instance,
                                #   not a field type class
                                #
                                # `ValueObject` instances are essentially treated as `Dict`. If not pickled,
                                #   they are persisted as JSON.
                                if isinstance(field_obj.content_type, ValueObject):
                                    if not field_obj.pickled:
                                        field_mapping_type = psql.JSON
                                    else:
                                        field_mapping_type = sa_types.PickleType
                                else:
                                    field_mapping_type = field_mapping.get(
                                        field_obj.content_type
                                    )
                                type_args.append(field_mapping_type)
                            else:
                                type_args.append(sa_types.Text)

                    # Default to the text type if no mapping is found
                    if not sa_type_cls:
                        sa_type_cls = sa_types.String

                    # Build the column arguments
                    col_args = {
                        "primary_key": field_obj.identifier,
                        "nullable": not field_obj.required,
                        "unique": field_obj.unique,
                    }

                    # Update the arguments based on the field type
                    if issubclass(field_cls, String):
                        type_kwargs["length"] = field_obj.max_length

                    # Update the attributes of the class
                    column = Column(sa_type_cls(*type_args, **type_kwargs), **col_args)
                    setattr(subclass, attribute_name, column)  # Set class attribute

        super().__init_subclass__(**kwargs)

    @orm.declared_attr
    def __tablename__(cls):
        return derive_schema_name(cls)

    @classmethod
    def from_entity(cls, entity):
        """Convert the entity to a model object"""
        item_dict = {}
        for attribute_obj in attributes(cls.meta_.entity_cls).values():
            if isinstance(attribute_obj, Reference):
                item_dict[attribute_obj.relation.attribute_name] = (
                    attribute_obj.relation.value
                )
            else:
                item_dict[attribute_obj.attribute_name] = getattr(
                    entity, attribute_obj.attribute_name
                )
        return cls(**item_dict)

    @classmethod
    def to_entity(cls, model_obj: "SqlalchemyModel"):
        """Convert the model object to an entity"""
        item_dict = {}
        for field_name in attributes(cls.meta_.entity_cls):
            item_dict[field_name] = getattr(model_obj, field_name, None)
        return cls.meta_.entity_cls(item_dict)


class SADAO(BaseDAO):
    """DAO implementation for Databases compliant with SQLAlchemy"""

    def __repr__(self) -> str:
        return f"SQLAlchemyDAO <{self.entity_cls.__name__}>"

    def _get_session(self):
        """Returns an active connection to the persistence store.

        - If there is an active transaction, the connection associated with the transaction (in the UoW) is returned
        - If the DAO has been explicitly instructed to work outside a UoW (with the help of `_outside_uow`), or if
            there are no active transactions, a new connection is retrieved from the provider and returned.

        Overridden here instead of using the version in `BaseDAO` because the connection needs to be started
            with a call to `begin()` if it is not yet active (checked with `is_active`)
        """
        if current_uow and not self._outside_uow:
            return current_uow.get_session(self.provider.name)
        else:
            new_connection = self.provider.get_connection()
            if not new_connection.is_active:
                new_connection.begin()
            return new_connection

    def _build_filters(self, criteria: Q):
        """Recursively Build the filters from the criteria object"""
        # Decide the function based on the connector type
        func = and_ if criteria.connector == criteria.AND else or_
        params = []
        for child in criteria.children:
            if isinstance(child, Q):
                # Call the function again with the child
                params.append(self._build_filters(child))
            else:
                # Find the lookup class and the key
                stripped_key, lookup_class = self.provider._extract_lookup(child[0])

                # Instantiate the lookup class and get the expression
                lookup = lookup_class(stripped_key, child[1], self.model_cls)
                if criteria.negated:
                    params.append(~lookup.as_expression())
                else:
                    params.append(lookup.as_expression())

        return func(*params)

    def _filter(
        self, criteria: Q, offset: int = 0, limit: int = 10, order_by: list = ()
    ) -> ResultSet:
        """Filter objects from the sqlalchemy database"""
        conn = self._get_session()
        qs = conn.query(self.model_cls)

        # Build the filters from the criteria
        if criteria.children:
            qs = qs.filter(self._build_filters(criteria))

        # Apply the order by clause if present
        order_cols = []
        for order_col in order_by:
            col = getattr(self.model_cls, order_col.lstrip("-"))
            if order_col.startswith("-"):
                order_cols.append(col.desc())
            else:
                order_cols.append(col)
        qs = qs.order_by(*order_cols)
        qs_without_limit = qs
        qs = qs.limit(limit).offset(offset)

        # Return the results
        try:
            items = qs.all()
            result = ResultSet(
                offset=offset, limit=limit, total=qs_without_limit.count(), items=items
            )
        except DatabaseError as exc:
            logger.error(f"Error while filtering: {exc}")
            raise
        finally:
            if not current_uow:
                conn.commit()
                conn.close()

        return result

    def _create(self, model_obj):
        """Add a new record to the sqlalchemy database"""
        conn = self._get_session()

        try:
            conn.add(model_obj)
        except DatabaseError as exc:
            logger.error(f"Error while creating: {exc}")
            raise
        finally:
            if not current_uow:
                conn.commit()
                conn.close()

        return model_obj

    def _update(self, model_obj):
        """Update a record in the sqlalchemy database"""
        conn = self._get_session()
        db_item = None

        # Fetch the record from database
        try:
            identifier = getattr(model_obj, id_field(self.entity_cls).attribute_name)
            db_item = conn.query(self.model_cls).get(
                identifier
            )  # This will raise exception if object was not found
        except DatabaseError as exc:
            logger.error(f"Database Record not found: {exc}")
            raise

        if db_item is None:
            conn.rollback()
            conn.close()
            raise ObjectNotFoundError(
                {
                    "_entity": f"`{self.entity_cls.__name__}` object with identifier {identifier} "
                    f"does not exist."
                }
            )

        # Sync DB Record with current changes. When the session is committed, changes are automatically synced
        try:
            for attribute in attributes(self.entity_cls):
                if attribute != id_field(self.entity_cls).attribute_name and getattr(
                    model_obj, attribute
                ) != getattr(db_item, attribute):
                    setattr(db_item, attribute, getattr(model_obj, attribute))
        except DatabaseError as exc:
            logger.error(f"Error while updating: {exc}")
            raise
        finally:
            if not current_uow:
                conn.commit()
                conn.close()

        return model_obj

    def _update_all(self, criteria: Q, *args, **kwargs):
        """Update all objects satisfying the criteria"""
        conn = self._get_session()
        qs = conn.query(self.model_cls).filter(self._build_filters(criteria))
        try:
            values = {}
            if args:
                values = args[
                    0
                ]  # `args[0]` is required because `*args` is sent as a tuple
            values.update(kwargs)
            updated_count = qs.update(values)
        except DatabaseError as exc:
            logger.error(f"Error while updating all: {exc}")
            raise
        finally:
            if not current_uow:
                conn.commit()
                conn.close()

        return updated_count

    def _delete(self, model_obj):
        """Delete the entity record in the dictionary"""
        conn = self._get_session()
        db_item = None

        # Fetch the record from database
        try:
            identifier = getattr(model_obj, id_field(self.entity_cls).attribute_name)
            db_item = conn.query(self.model_cls).get(
                identifier
            )  # This will raise exception if object was not found
        except DatabaseError as exc:
            logger.error(f"Database Record not found: {exc}")
            raise

        if db_item is None:
            conn.rollback()
            conn.close()
            raise ObjectNotFoundError(
                {
                    "_entity": f"`{self.entity_cls.__name__}` object with identifier {identifier} "
                    f"does not exist."
                }
            )

        try:
            conn.delete(db_item)
        except DatabaseError as exc:
            logger.error(f"Error while deleting: {exc}")
            raise
        finally:
            if not current_uow:
                conn.commit()
                conn.close()

        return model_obj

    def _delete_all(self, criteria: Q = None):
        """Delete a record from the sqlalchemy database"""
        conn = self._get_session()

        del_count = 0
        if criteria:
            qs = conn.query(self.model_cls).filter(self._build_filters(criteria))
        else:
            qs = conn.query(self.model_cls)

        try:
            del_count = qs.delete()
        except DatabaseError as exc:
            logger.error(f"Error while deleting all: {exc}")
            raise
        finally:
            if not current_uow:
                conn.commit()
                conn.close()

        return del_count

    def _raw(self, query: Any, data: Any = None):
        """Run a raw query on the repository and return entity objects"""
        assert isinstance(query, str)

        conn = self._get_session()
        try:
            results = conn.execute(text(query))

            entity_items = []
            for item in results:
                entity = self.model_cls.to_entity(item)
                entity.state_.mark_retrieved()
                entity_items.append(entity)

            result = ResultSet(
                offset=0,
                limit=len(entity_items),
                total=len(entity_items),
                items=entity_items,
            )
        except DatabaseError as exc:
            logger.error(f"Error while running raw query: {exc}")
            raise
        finally:
            if not current_uow:
                conn.commit()
                conn.close()

        return result


class SAProvider(BaseProvider):
    """Provider Implementation class for SQLAlchemy"""

    class databases(Enum):
        postgresql = "postgresql"
        sqlite = "sqlite"

    def __init__(self, *args, **kwargs):
        """Initialize and maintain Engine"""
        # Since SQLAlchemyProvider can cater to multiple databases, it is important
        #   that we know which database we are dealing with, to run database-specific
        #   statements like `PRAGMA` for SQLite.
        if "database" not in args[2]:
            logger.error(f"Missing `database` information in conn_info: {args[2]}")
            raise ConfigurationError("Missing `database` attribute in Connection info")

        super().__init__(*args, **kwargs)

        kwargs = self._get_database_specific_engine_args()

        self._engine = create_engine(
            make_url(self.conn_info["database_uri"]),
            json_serializer=_custom_json_dumps,
            **kwargs,
        )

        if self.conn_info["database"] == self.databases.postgresql.value:
            # Nest database tables under a schema, so that we have complete control
            #   on creating/dropping db structures. We cannot control structures in the
            #   the default `public` schema.
            #
            # Use `SCHEMA` value if specified as part of the conn info. Otherwise, construct
            #   and use default schema name as `DB`_schema.
            schema = (
                self.conn_info["SCHEMA"] if "SCHEMA" in self.conn_info else "public"
            )

            self._metadata = MetaData(schema=schema)
        else:
            self._metadata = MetaData()

        # A temporary cache of already constructed model classes
        self._model_classes = {}

    def _get_database_specific_engine_args(self):
        """Supplies additional database-specific arguments to SQLAlchemy Engine.

        Return: a dictionary with database-specific SQLAlchemy Engine arguments.
        """
        if self.conn_info["database"] == self.databases.postgresql.value:
            return {"isolation_level": "AUTOCOMMIT"}

        return {}

    def _get_database_specific_session_args(self):
        """Set Database specific session parameters.

        Depending on the database in use, this method supplies
        additional arguments while constructing sessions.

        Return: a dictionary with additional arguments and values.
        """
        if self.conn_info["database"] == self.databases.postgresql.value:
            return {"autoflush": False}

        return {}

    def get_session(self):
        """Establish a session to the Database"""
        # Create the session
        kwargs = self._get_database_specific_session_args()
        session_factory = orm.sessionmaker(
            bind=self._engine, expire_on_commit=False, **kwargs
        )
        session_cls = orm.scoped_session(session_factory)

        return session_cls

    def _execute_database_specific_connection_statements(self, conn):
        """Execute connection statements depending on the database in use.

        Each database has a unique set of commands and associated format to control
        connection-related parameters. Since we use SQLAlchemy, statements should
        be run dynamically based on the database in use.

        Arguments:
        * conn: An active connection object to the database

        Return: None
        """
        if self.conn_info["database"] == self.databases.sqlite.value:
            conn.execute(text("PRAGMA case_sensitive_like = ON;"))

        return conn

    def get_connection(self, session_cls=None):
        """Create the connection to the Database instance"""
        # If this connection has to be created within an existing session,
        #   ``session_cls`` will be provided as an argument.
        #   Otherwise, fetch a new ``session_cls`` from ``get_session()``
        if session_cls is None:
            session_cls = self.get_session()

        conn = session_cls()
        conn = self._execute_database_specific_connection_statements(conn)

        return conn

    def _data_reset(self):
        conn = self._engine.connect()

        transaction = conn.begin()

        if self.conn_info["database"] == self.databases.sqlite.value:
            conn.execute(text("PRAGMA foreign_keys = OFF;"))

        for table in self._metadata.sorted_tables:
            conn.execute(table.delete())

        if self.conn_info["database"] == self.databases.sqlite.value:
            conn.execute(text("PRAGMA foreign_keys = ON;"))

        transaction.commit()

        # Discard any active Unit of Work
        if current_uow and current_uow.in_progress:
            current_uow.rollback()

    def _create_database_artifacts(self):
        for _, aggregate_record in self.domain.registry.aggregates.items():
            self.domain.repository_for(aggregate_record.cls)._dao

        self._metadata.create_all(self._engine)

    def _drop_database_artifacts(self):
        self._metadata.drop_all(self._engine)
        self._metadata.clear()

    def decorate_model_class(self, entity_cls, model_cls):
        schema_name = derive_schema_name(model_cls)

        # Return the model class if it was already seen/decorated
        if schema_name in self._model_classes:
            return self._model_classes[schema_name]

        # If `model_cls` is already subclassed from SqlAlchemyModel,
        #   this method call is a no-op
        if issubclass(model_cls, SqlalchemyModel):
            return model_cls
        else:
            # Strip out `Column` attributes from the model class
            # Create a deep copy to make this work
            # https://stackoverflow.com/a/62528033/1858466
            columns = copy.deepcopy(
                {
                    key: value
                    for key, value in vars(model_cls).items()
                    if isinstance(value, Column)
                }
            )

            custom_attrs = {
                key: value
                for (key, value) in vars(model_cls).items()
                if key not in ["Meta", "__module__", "__doc__", "__weakref__"]
                and not isinstance(value, Column)
            }

            # Add the earlier copied columns to the custom attributes
            custom_attrs = {**custom_attrs, **columns}

            meta_ = Options(model_cls.meta_)
            meta_.entity_cls = entity_cls
            meta_.schema_name = (
                schema_name if meta_.schema_name is None else meta_.schema_name
            )

            custom_attrs.update(
                {"meta_": meta_, "engine": self._engine, "metadata": self._metadata}
            )
            # FIXME Ensure the custom model attributes are constructed properly
            decorated_model_cls = type(
                model_cls.__name__, (SqlalchemyModel, model_cls), custom_attrs
            )

            # Memoize the constructed model class
            self._model_classes[schema_name] = decorated_model_cls

            return decorated_model_cls

    def construct_model_class(self, entity_cls):
        """Return a fully-baked Model class for a given Entity class"""
        model_cls = None

        # Return the model class if it was already seen/decorated
        if entity_cls.meta_.schema_name in self._model_classes:
            model_cls = self._model_classes[entity_cls.meta_.schema_name]
        else:
            # Construct a new Meta object with existing values
            meta_ = Options()
            meta_.entity_cls = entity_cls
            # If schema_name is not provided, sqlalchemy can throw
            #   sqlalchemy.exc.InvalidRequestError: Class does not
            #   have a __table__ or __tablename__ specified and
            #   does not inherit from an existing table-mapped class
            meta_.schema_name = entity_cls.meta_.schema_name

            attrs = {
                "meta_": meta_,
                "engine": self._engine,
                "metadata": self._metadata,
            }
            # FIXME Ensure the custom model attributes are constructed properly
            model_cls = type(entity_cls.__name__ + "Model", (SqlalchemyModel,), attrs)

            # Memoize the constructed model class
            self._model_classes[entity_cls.meta_.schema_name] = model_cls

        # Set Entity Class as a class level attribute for the Model, to be able to reference later.
        return model_cls

    def get_dao(self, entity_cls, model_cls):
        """Return a DAO object configured with a live connection"""
        return SADAO(self.domain, self, entity_cls, model_cls)

    def raw(self, query: Any, data: Any = None):
        """Run raw query on Provider"""
        if data is None:
            data = {}
        assert isinstance(query, str)
        assert isinstance(data, (dict, None))

        return self.get_connection().execute(text(query), data)


operators = {
    "exact": "__eq__",
    "iexact": "ilike",
    "contains": "contains",
    "icontains": "ilike",
    "startswith": "startswith",
    "endswith": "endswith",
    "gt": "__gt__",
    "gte": "__ge__",
    "lt": "__lt__",
    "lte": "__le__",
    "in": "in_",
    "any": "any",
    "overlap": "overlap",
}


class DefaultLookup(BaseLookup):
    """Base class with default implementation of expression construction"""

    def __init__(self, source, target, model_cls):
        """Source is LHS and Target is RHS of a comparsion"""
        self.model_cls = model_cls
        super().__init__(source, target)

    def process_source(self):
        """Return source with transformations, if any"""
        source_col = getattr(self.model_cls, self.source)
        return source_col

    def process_target(self):
        """Return target with transformations, if any"""
        return self.target

    def as_expression(self):
        lookup_func = getattr(self.process_source(), operators[self.lookup_name])
        return lookup_func(self.process_target())


@SAProvider.register_lookup
class Exact(DefaultLookup):
    """Exact Match Query"""

    lookup_name = "exact"


@SAProvider.register_lookup
class IExact(DefaultLookup):
    """Exact Case-Insensitive Match Query"""

    lookup_name = "iexact"


@SAProvider.register_lookup
class Contains(DefaultLookup):
    """Exact Contains Query"""

    lookup_name = "contains"


@SAProvider.register_lookup
class IContains(DefaultLookup):
    """Case-Insensitive Contains Query"""

    lookup_name = "icontains"

    def process_target(self):
        """Return target in lowercase"""
        assert isinstance(self.target, str)
        return f"%{super().process_target()}%"


@SAProvider.register_lookup
class Startswith(DefaultLookup):
    """Exact Contains Query"""

    lookup_name = "startswith"


@SAProvider.register_lookup
class Endswith(DefaultLookup):
    """Exact Contains Query"""

    lookup_name = "endswith"


@SAProvider.register_lookup
class GreaterThan(DefaultLookup):
    """Greater than Query"""

    lookup_name = "gt"


@SAProvider.register_lookup
class GreaterThanOrEqual(DefaultLookup):
    """Greater than or Equal Query"""

    lookup_name = "gte"


@SAProvider.register_lookup
class LessThan(DefaultLookup):
    """Less than Query"""

    lookup_name = "lt"


@SAProvider.register_lookup
class LessThanOrEqual(DefaultLookup):
    """Less than or Equal Query"""

    lookup_name = "lte"


@SAProvider.register_lookup
class In(DefaultLookup):
    """In Query"""

    lookup_name = "in"

    def process_target(self):
        """Ensure target is a list or tuple"""
        assert isinstance(self.target, (list, tuple))
        return super().process_target()


@SAProvider.register_lookup
class Any(DefaultLookup):
    """Any Query"""

    lookup_name = "any"


@SAProvider.register_lookup
class Overlap(DefaultLookup):
    """Overlap Query"""

    lookup_name = "overlap"



import importlib
import logging
from collections import defaultdict
from typing import List, Optional, Type

from protean import BaseEvent, BaseEventHandler
from protean.core.command import BaseCommand
from protean.core.command_handler import BaseCommandHandler
from protean.core.event_sourced_repository import (
    BaseEventSourcedRepository,
    event_sourced_repository_factory,
)
from protean.exceptions import ConfigurationError, NotSupportedError
from protean.utils import fqn
from protean.utils.mixins import Message

logger = logging.getLogger(__name__)

EVENT_STORE_PROVIDERS = {
    "memory": "protean.adapters.event_store.memory.MemoryEventStore",
    "message_db": "protean.adapters.event_store.message_db.MessageDBStore",
}


class EventStore:
    def __init__(self, domain):
        self.domain = domain
        self._event_store = None
        self._event_streams = None
        self._command_streams = None

    @property
    def store(self):
        if self._event_store is None:
            self._initialize()

        return self._event_store

    def _initialize(self):
        if not self._event_store:
            logger.debug("Initializing Event Store...")

            configured_event_store = self.domain.config["event_store"]
            if configured_event_store and isinstance(configured_event_store, dict):
                event_store_full_path = EVENT_STORE_PROVIDERS[
                    configured_event_store["provider"]
                ]
                event_store_module, event_store_class = event_store_full_path.rsplit(
                    ".", maxsplit=1
                )

                event_store_cls = getattr(
                    importlib.import_module(event_store_module), event_store_class
                )

                store = event_store_cls(self.domain, configured_event_store)
            else:
                raise ConfigurationError(
                    "Configure at least one event store in the domain"
                )

            self._event_store = store

            self._initialize_event_streams()
            self._initialize_command_streams()

        return self._event_store

    def _initialize_event_streams(self):
        self._event_streams = defaultdict(set)

        for _, record in self.domain.registry.event_handlers.items():
            stream_name = (
                record.cls.meta_.stream_name
                or record.cls.meta_.part_of.meta_.stream_name
            )
            self._event_streams[stream_name].add(record.cls)

    def _initialize_command_streams(self):
        self._command_streams = defaultdict(set)

        for _, record in self.domain.registry.command_handlers.items():
            self._command_streams[record.cls.meta_.part_of.meta_.stream_name].add(
                record.cls
            )

    def repository_for(self, part_of):
        if self._event_store is None:
            self._initialize()

        repository_cls = type(
            part_of.__name__ + "Repository", (BaseEventSourcedRepository,), {}
        )
        repository_cls = event_sourced_repository_factory(
            repository_cls, part_of=part_of
        )
        return repository_cls(self.domain)

    def handlers_for(self, event: BaseEvent) -> List[BaseEventHandler]:
        if self._event_streams is None:
            self._initialize_event_streams()

        all_stream_handlers = self._event_streams.get("$all", set())

        # Take the Aggregate's stream_name
        aggregate_stream_name = None
        if event.meta_.aggregate_cluster:
            aggregate_stream_name = event.meta_.aggregate_cluster.meta_.stream_name

        stream_name = event.meta_.stream_name or aggregate_stream_name
        stream_handlers = self._event_streams.get(stream_name, set())

        return set.union(stream_handlers, all_stream_handlers)

    def command_handler_for(self, command: BaseCommand) -> Optional[BaseCommandHandler]:
        if self._command_streams is None:
            self._initialize_command_streams()

        stream_name = command.meta_.stream_name or (
            command.meta_.part_of.meta_.stream_name if command.meta_.part_of else None
        )

        if not stream_name:
            return None

        handler_classes = self._command_streams.get(stream_name, set())

        # No command handlers have been configured to run this command
        if len(handler_classes) == 0:
            return None

        # Ensure that a command has a unique handler across all handlers
        # FIXME Perform this check on domain spin-up?
        handler_methods = set()
        for handler_cls in handler_classes:
            try:
                handler_method = next(
                    iter(handler_cls._handlers[fqn(command.__class__)])
                )
                handler_methods.add((handler_cls, handler_method))
            except StopIteration:
                pass

        if len(handler_methods) > 1:
            raise NotSupportedError(
                f"Command {command.__class__.__name__} cannot be handled by multiple handlers"
            )

        return next(iter(handler_methods))[0] if handler_methods else None

    def last_event_of_type(
        self, event_cls: Type[BaseEvent], stream_name: str = None
    ) -> BaseEvent:
        stream_name = stream_name or "$all"
        events = [
            event
            for event in self.domain.event_store.store._read(stream_name)
            if event["type"] == fqn(event_cls)
        ]

        return Message.from_dict(events[-1]).to_object() if len(events) > 0 else None

    def events_of_type(
        self, event_cls: Type[BaseEvent], stream_name: str = None
    ) -> List[BaseEvent]:
        """Read events of a specific type in a given stream.

        This is a utility method, especially useful for testing purposes, that retrives events of a
        specific type from the event store.

        If no stream is specified, events of the requested type will be retrieved from all streams.

        :param event_cls: Class of the event type to be retrieved
        :param stream_name: Stream from which events are to be retrieved
        :type event_cls: BaseEvent Class
        :type stream_name: String, optional, default is `None`
        :return: A list of events of `event_cls` type
        :rtype: list
        """
        stream_name = stream_name or "$all"
        return [
            Message.from_dict(event).to_object()
            for event in self.domain.event_store.store._read(stream_name)
            if event["type"] == fqn(event_cls)
        ]



import logging

from protean import BaseEventSourcedAggregate
from protean.container import Element, OptionsMixin
from protean.exceptions import (
    IncorrectUsageError,
    NotSupportedError,
    ObjectNotFoundError,
)
from protean.fields import Identifier
from protean.globals import current_uow
from protean.utils import DomainObjects, derive_element_class

logger = logging.getLogger(__name__)


class BaseEventSourcedRepository(Element, OptionsMixin):
    element_type = DomainObjects.EVENT_SOURCED_REPOSITORY

    @classmethod
    def _default_options(cls):
        return [("part_of", None)]

    def __new__(cls, *args, **kwargs):
        # Prevent instantiation of `BaseEventSourcedRepository itself`
        if cls is BaseEventSourcedRepository:
            raise NotSupportedError("BaseEventSourcedRepository cannot be instantiated")
        return super().__new__(cls)

    def __init__(self, domain) -> None:
        self._domain = domain

    def add(self, aggregate: BaseEventSourcedAggregate) -> None:
        current_uow._add_to_identity_map(aggregate)

    def get(self, identifier: Identifier) -> BaseEventSourcedAggregate:
        """Retrieve a fully-formed Aggregate from a stream of Events.

        If the aggregate was already loaded in the current UnitOfWork,
        `get` will return the aggregate object instead of loading it again
        from events.

        Args:
            identifier (Identifier): Aggregate identifier

        Raises:
            ObjectNotFoundError: When no stream with identifier is found

        Returns:
            BaseEventSourcedAggregate: The fully-loaded aggregate object
        """
        # Return aggregate if it was already loaded and is present in current
        #   UnitOfWork's identity map.
        if current_uow and identifier in current_uow._identity_map:
            return current_uow._identity_map[identifier]

        aggregate = self._domain.event_store.store.load_aggregate(
            self.meta_.part_of, identifier
        )

        if not aggregate:
            raise ObjectNotFoundError(
                {
                    "_entity": f"`{self.meta_.part_of.__name__}` object with identifier {identifier} "
                    f"does not exist."
                }
            )

        return aggregate


def event_sourced_repository_factory(element_cls, **opts):
    element_cls = derive_element_class(element_cls, BaseEventSourcedRepository, **opts)

    if not element_cls.meta_.part_of:
        raise IncorrectUsageError(
            {
                "_entity": [
                    f"Repository `{element_cls.__name__}` should be associated with an Aggregate"
                ]
            }
        )

    if not issubclass(element_cls.meta_.part_of, BaseEventSourcedAggregate):
        raise IncorrectUsageError(
            {
                "_entity": [
                    f"Repository `{element_cls.__name__}` can only be associated with an Aggregate"
                ]
            }
        )

    return element_cls



import collections
import re
import time
from threading import RLock
from typing import Optional, Union

from protean.core.view import BaseView
from protean.port.cache import BaseCache
from protean.reflection import id_field
from protean.utils.inflection import underscore


class TTLDict(collections.abc.MutableMapping):
    def __init__(self, default_ttl, *args, **kwargs):
        self._default_ttl = default_ttl
        self._values = {}
        self._lock = RLock()
        self.update(*args, **kwargs)

    def __repr__(self):
        return "<TTLDict@%#08x; ttl=%r, v=%r;>" % (
            id(self),
            self._default_ttl,
            self._values,
        )

    def set_ttl(self, key, ttl, now=None):
        """Set TTL for the given key"""
        if now is None:
            now = time.time()
        with self._lock:
            _expire, value = self._values[key]
            self._values[key] = (now + ttl, value)

    def get_ttl(self, key, now=None):
        """Return remaining TTL for a key"""
        if now is None:
            now = time.time()
        with self._lock:
            expire, _value = self._values[key]
            return expire - now

    def expire_at(self, key, timestamp):
        """Set the key expire timestamp"""
        with self._lock:
            _expire, value = self._values[key]
            self._values[key] = (timestamp, value)

    def is_expired(self, key, now=None, remove=False):
        """Check if key has expired"""
        with self._lock:
            if now is None:
                now = time.time()
            expire, _value = self._values[key]
            if expire is None:
                return False
            expired = expire < now
            if expired and remove:
                self.__delitem__(key)
            return expired

    def __len__(self):
        with self._lock:
            for key in self._values.keys():
                self.is_expired(key, remove=True)
            return len(self._values)

    def __iter__(self):
        with self._lock:
            for key in self._values.keys():
                if not self.is_expired(key):
                    yield key

    def __setitem__(self, key, value):
        with self._lock:
            if self._default_ttl is None:
                expire = None
            else:
                expire = time.time() + self._default_ttl
            self._values[key] = (expire, value)

    def __delitem__(self, key):
        with self._lock:
            del self._values[key]

    def __getitem__(self, key):
        with self._lock:
            self.is_expired(key, remove=True)
            return self._values[key][1]


class MemoryCache(BaseCache):
    def __init__(self, name, domain, conn_info: dict):
        """Initialize Cache with Connection/Adapter details"""

        # In case of `MemoryCache`, the `cache` value will always be `memory`.
        conn_info["cache"] = "memory"
        super().__init__(name, domain, conn_info)

        # The Data Cache
        self._db = TTLDict(self.conn_info.get("TTL") or 300)

        self._lock = RLock()

    def ping(self):
        """Always returns True for memory cache"""
        return True

    def get_connection(self):
        """Get the connection object for the repository"""
        return self._db._values

    def add(self, view: BaseView, ttl: Optional[Union[int, float]] = None) -> None:
        """Add view record to cache

        KEY: View ID
        Value: View Data (derived from `to_dict()`)

        TTL is in seconds.

        Args:
            view (BaseView): View Instance containing data
            ttl (int, float, optional): Timeout in seconds. Defaults to None.
        """
        identifier = getattr(view, id_field(view).field_name)
        key = f"{underscore(view.__class__.__name__)}:::{identifier}"

        self._db[key] = view.to_dict()

        if ttl:
            self._db.set_ttl(key, ttl)

    def get(self, key):
        view_name = key.split(":::")[0]
        view_cls = self._views[view_name]

        value = self._db.get(key)
        return view_cls(value) if value else None

    def get_all(self, key_pattern, last_position=0, size=25):
        # FIXME Handle Pagination with Last Position
        # FIXME Handle Pagination with Size
        view_name = key_pattern.split(":::")[0]
        view_cls = self._views[view_name]

        key_list = self._db.keys()
        regex = re.compile(key_pattern)
        results = list(filter(regex.match, key_list))

        return [view_cls(self._db.get(key)) for key in results]

    def count(self, key_pattern):
        key_list = self._db.keys()
        regex = re.compile(key_pattern)
        return len(list(filter(regex.match, key_list)))

    def remove(self, view):
        identifier = getattr(view, id_field(view).field_name)
        key = f"{underscore(view.__class__.__name__)}:::{identifier}"
        del self._db[key]

    def remove_by_key(self, key):
        del self._db[key]

    def remove_by_key_pattern(self, key_pattern):
        full_key_list = self._db.keys()
        regex = re.compile(key_pattern)
        keys_to_delete = list(filter(regex.match, full_key_list))
        for key in keys_to_delete:
            del self._db[key]

    def flush_all(self):
        self._db = {}

    def set_ttl(self, key, ttl):
        self._db.set_ttl(key, ttl)

    def get_ttl(self, key):
        return self._db.get_ttl(key)



from abc import abstractmethod

from protean.container import BaseContainer, OptionsMixin
from protean.fields import Dict, List, String, Text
from protean.utils import (
    DomainObjects,
    convert_str_values_to_list,
    derive_element_class,
)


class BaseEmailProvider:
    """
    Base class for email backend implementations.

    Concrete implementations must overwrite `send_email()`.
    ```
    """

    def __init__(self, name, domain, conn_info, fail_silently=False, **kwargs):
        self.name = name
        self.domain = domain
        self.conn_info = conn_info
        self.fail_silently = fail_silently

    @abstractmethod
    def send_email(self, email_message):
        """
        Send EmailMessage object via registered email provider.
        """
        raise NotImplementedError(
            "Concrete implementations of BaseEmailBackend "
            "must override send_email() method"
        )


class BaseEmail(BaseContainer, OptionsMixin):  # FIXME Remove OptionsMixin
    """Base Email class that should implemented by all Domain Email Messages.

    This is also a marker class that is referenced when emails are registered
    with the domain.
    """

    element_type = DomainObjects.EMAIL

    @classmethod
    def _default_options(cls):
        return [("provider", "default")]

    subject = String()
    from_email = String()
    to = List(content_type=String)
    bcc = List(content_type=String)
    cc = List(content_type=String)
    reply_to = String()

    # Supplied content
    text = Text()
    html = Text()

    # JSON data with template
    data = Dict()
    template = String()

    def defaults(self):
        """
        Initialize a single email message (which can be sent to multiple
        recipients).
        """
        self.to = convert_str_values_to_list(self.to)
        self.cc = convert_str_values_to_list(self.cc)
        self.bcc = convert_str_values_to_list(self.bcc)
        self.reply_to = (
            convert_str_values_to_list(self.reply_to)
            if self.reply_to
            else self.from_email
        )

    @property
    def recipients(self):
        """
        Return a list of all recipients of the email (includes direct
        addressees as well as Cc and Bcc entries).
        """
        return [email for email in (self.to + self.cc + self.bcc) if email]


def email_factory(element_cls, **kwargs):
    return derive_element_class(element_cls, BaseEmail)



"""Aggregate Functionality and Classes"""

import inspect
import logging

from protean.core.entity import BaseEntity
from protean.exceptions import NotSupportedError
from protean.fields import Integer
from protean.utils import DomainObjects, derive_element_class, inflection

logger = logging.getLogger(__name__)


class BaseAggregate(BaseEntity):
    """This is the base class for Domain Aggregates.

    Aggregates are fundamental, coarse-grained building blocks of a domain model. They are
    conceptual wholes - they enclose all behaviors and data of a distinct domain concept.
    Aggregates are often composed of one or more Aggregate Elements (Entities and Value Objests),
    that work together to codify a concept.

    This class provides helper methods to custom define aggregate attributes, and query attribute
    names during runtime.

    Basic Usage::

        @domain.aggregate
        class Dog:
            id = field.Integer(identifier=True)
            name = field.String(required=True, max_length=50)
            age = field.Integer(default=5)
            owner = field.String(required=True, max_length=15)

    During persistence, the model associated with this entity is retrieved dynamically from
        the repository factory. A model object is usually pre-initialized with a live DB connection.
    """

    element_type = DomainObjects.AGGREGATE

    def __new__(cls, *args, **kwargs):
        if cls is BaseAggregate:
            raise NotSupportedError("BaseAggregate cannot be instantiated")
        return super().__new__(cls)

    # Track current version of Aggregate
    _version = Integer(default=-1)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Set root in all child elements
        #   This is where we kick-off the process of setting the owner and root
        self._set_root_and_owner(self, self)

    @classmethod
    def _default_options(cls):
        return [
            ("auto_add_id_field", True),
            ("provider", "default"),
            ("aggregate_cluster", None),
            ("model", None),
            ("stream_name", inflection.underscore(cls.__name__)),
            ("schema_name", inflection.underscore(cls.__name__)),
        ]


def aggregate_factory(element_cls, **kwargs):
    element_cls = derive_element_class(element_cls, BaseAggregate, **kwargs)

    # Iterate through methods marked as `@invariant` and record them for later use
    #   `_invariants` is a dictionary initialized in BaseEntity.__init_subclass__
    methods = inspect.getmembers(element_cls, predicate=inspect.isroutine)
    for method_name, method in methods:
        if not (
            method_name.startswith("__") and method_name.endswith("__")
        ) and hasattr(method, "_invariant"):
            element_cls._invariants[method._invariant][method_name] = method

    return element_cls


# Context manager to temporarily disable invariant checks on aggregate
class atomic_change:
    def __init__(self, aggregate):
        self.aggregate = aggregate

    def __enter__(self):
        # Temporary disable invariant checks
        self.aggregate._precheck()
        self.aggregate._disable_invariant_checks = True

    def __exit__(self, *args):
        # Validate on exit to trigger invariant checks
        self.aggregate._disable_invariant_checks = False
        self.aggregate._postcheck()



import inspect
import logging

from protean.container import Element, OptionsMixin
from protean.core.event import BaseEvent
from protean.exceptions import IncorrectUsageError, NotSupportedError
from protean.utils import DomainObjects, derive_element_class, fully_qualified_name
from protean.utils.mixins import HandlerMixin

logger = logging.getLogger(__name__)


class BaseEventHandler(Element, HandlerMixin, OptionsMixin):
    """Base Event Handler to be inherited by all event handlers"""

    element_type = DomainObjects.EVENT_HANDLER

    def __new__(cls, *args, **kwargs):
        if cls is BaseEventHandler:
            raise NotSupportedError("BaseEventHandler cannot be instantiated")
        return super().__new__(cls)

    @classmethod
    def _default_options(cls):
        part_of = (
            getattr(cls.meta_, "part_of") if hasattr(cls.meta_, "part_of") else None
        )

        return [
            ("part_of", None),
            ("stream_name", part_of.meta_.stream_name if part_of else None),
            ("source_stream", None),
        ]


def event_handler_factory(element_cls, **opts):
    element_cls = derive_element_class(element_cls, BaseEventHandler, **opts)

    if not (element_cls.meta_.part_of or element_cls.meta_.stream_name):
        raise IncorrectUsageError(
            {
                "_entity": [
                    f"Event Handler `{element_cls.__name__}` needs to be associated with an aggregate or a stream"
                ]
            }
        )

    # Iterate through methods marked as `@handle` and construct a handler map
    #
    # Also, if `_target_cls` is an event, associate it with the event handler's
    #   aggregate or stream
    methods = inspect.getmembers(element_cls, predicate=inspect.isroutine)
    for method_name, method in methods:
        if not (
            method_name.startswith("__") and method_name.endswith("__")
        ) and hasattr(method, "_target_cls"):
            # `_handlers` is a dictionary mapping the event to the handler method.
            if method._target_cls == "$any":
                # This replaces any existing `$any` handler, by design. An Event Handler
                # can have only one `$any` handler method.
                element_cls._handlers["$any"] = {method}
            else:
                element_cls._handlers[fully_qualified_name(method._target_cls)].add(
                    method
                )

            # Associate Event with the handler's stream
            if inspect.isclass(method._target_cls) and issubclass(
                method._target_cls, BaseEvent
            ):
                # Order of preference:
                #   1. Stream name defined in event
                #   2. Stream name defined for the event handler
                #   3. Stream name derived from aggregate
                stream_name = element_cls.meta_.stream_name or (
                    element_cls.meta_.part_of.meta_.stream_name
                    if element_cls.meta_.part_of
                    else None
                )
                method._target_cls.meta_.stream_name = (
                    method._target_cls.meta_.stream_name or stream_name
                )

    return element_cls



import logging
import os
import re

import tomllib

from protean.exceptions import ConfigurationError

logger = logging.getLogger(__name__)


class ConfigAttribute:
    """Makes an attribute forward to the config"""

    def __init__(self, name):
        self.__name__ = name

    def __get__(self, obj, type=None):
        return obj.config[self.__name__]

    def __set__(self, obj, value):
        obj.config[self.__name__] = value


class Config2(dict):
    ENV_VAR_PATTERN = re.compile(r"\$\{([^}]+)\}")

    @classmethod
    def load_from_dict(cls, config: dict):
        """Load configuration from a dictionary."""
        return cls(**config)

    @classmethod
    def load_from_path(cls, path: str, defaults: dict = None):
        def find_config_file(directory: str):
            config_files = [".domain.toml", "domain.toml", "pyproject.toml"]
            for config_file in config_files:
                config_file_path = os.path.join(directory, config_file)
                if os.path.exists(config_file_path):
                    return config_file_path
            return None

        # Start checking from the provided path up to 2 parent directories
        current_dir = os.path.abspath(os.path.dirname(path))
        config_file_name = None

        for _ in range(3):  # Check the current directory and up to 2 parent directories
            config_file_name = find_config_file(current_dir)
            if config_file_name:
                break

            current_dir = os.path.dirname(current_dir)  # Move to the parent directory

        if not config_file_name:
            raise ConfigurationError(
                f"No configuration file found in {os.path.dirname(path)}"
            )

        config = {}
        if config_file_name:
            with open(config_file_name, "rb") as f:
                config = tomllib.load(f)

                # If pyproject.toml, extract protean configuration
                #   from the 'tool.protean' section
                if config_file_name.endswith("pyproject.toml"):
                    config = config.get("tool", {}).get("protean", {})

        # Merge with defaults
        config = cls._deep_merge(defaults, config)

        # Load environment variables
        config = cls._load_env_vars(config)

        return cls(**config)

    @classmethod
    def _deep_merge(cls, dict1: dict, dict2: dict):
        result = dict1.copy()
        for key, value in dict2.items():
            if (
                key in result
                and isinstance(result[key], dict)
                and isinstance(value, dict)
            ):
                result[key] = cls._deep_merge(result[key], value)
            else:
                result[key] = value
        return result

    @classmethod
    def _load_env_vars(cls, config):
        if isinstance(config, dict):
            for key, value in config.items():
                if isinstance(value, str):
                    config[key] = cls._replace_env_var(value)
                elif isinstance(value, dict):
                    config[key] = cls._load_env_vars(value)
                elif isinstance(value, list):
                    config[key] = [
                        cls._replace_env_var(item) if isinstance(item, str) else item
                        for item in value
                    ]
        return config

    @classmethod
    def _replace_env_var(cls, value):
        """Replace environment variables in a string.

        Cases:
        1. String does not have an environment variable. E.g. "attr-value" - Use as is
        2. String has an environment variable. E.g. "${ENV_VAR}" - Replace with value
        3. String has an environment variable with a default value. E.g. "${ENV_VAR|default-value}"
            - Replace with value or default value
        4. String has multiple environment variables. E.g. "${ENV_VAR1|default-value1} ${ENV_VAR2|default-value2}"
            - Replace all environment variables
        5. String has a mix of environment variables and static values. E.g. "attr-${ENV_VAR1|default-value1}"
            - Replace all environment variables
        """

        match = cls.ENV_VAR_PATTERN.search(value)
        while match:
            matched_string = match.group(1)

            if "|" in matched_string:
                # Default value provided
                env_var, default_value = matched_string.split("|", 1)
                env_value = os.getenv(env_var, default_value)
            else:
                # No default value provided
                env_value = os.getenv(matched_string)

            if env_value is None:
                raise ConfigurationError(
                    f"Environment variable {matched_string} is not set"
                )

            value = value.replace(f"${{{matched_string}}}", env_value)
            match = cls.ENV_VAR_PATTERN.search(value)

        return value



import asyncio
import logging
from typing import List, Union

from protean import BaseCommandHandler, BaseEventHandler
from protean.port.event_store import BaseEventStore
from protean.utils.mixins import Message, MessageType

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s,%(msecs)d %(levelname)s: %(message)s",
    datefmt="%H:%M:%S",
)

logger = logging.getLogger(__name__)


class Subscription:
    """
    Represents a subscription to a stream in the Protean event-driven architecture.

    A subscription allows a subscriber to receive and process messages from a specific stream.
    It provides methods to start and stop the subscription, as well as process messages in batches.
    """

    def __init__(
        self,
        engine,
        subscriber_id: str,
        stream_name: str,
        handler: Union[BaseEventHandler, BaseCommandHandler],
        messages_per_tick: int = 10,
        position_update_interval: int = 10,
        origin_stream_name: str | None = None,
        tick_interval: int = 1,
    ) -> None:
        """
        Initialize the Subscription object.

        Args:
            engine: The Protean engine instance.
            subscriber_id (str): The unique identifier for the subscriber.
            stream_name (str): The name of the stream to subscribe to.
            handler (Union[BaseEventHandler, BaseCommandHandler]): The event or command handler.
            messages_per_tick (int, optional): The number of messages to process per tick. Defaults to 10.
            position_update_interval (int, optional): The interval at which to update the current position. Defaults to 10.
            origin_stream_name (str | None, optional): The name of the origin stream to filter messages. Defaults to None.
            tick_interval (int, optional): The interval between ticks. Defaults to 1.
        """
        self.engine = engine

        self.store: BaseEventStore = engine.domain.event_store.store
        self.loop = engine.loop

        self.subscriber_id = subscriber_id
        self.stream_name = stream_name
        self.handler = handler
        self.messages_per_tick = messages_per_tick
        self.position_update_interval = position_update_interval
        self.origin_stream_name = origin_stream_name
        self.tick_interval = tick_interval

        self.subscriber_stream_name = f"position-${subscriber_id}"

        self.current_position: int = -1
        self.messages_since_last_position_write: int = 0

        self.keep_going = True  # Initially set to keep going

    async def start(self) -> None:
        """
        Start the subscription.

        This method initializes the subscription by loading the last position from the event store
        and starting the polling loop.

        Returns:
            None
        """
        logger.debug(f"Starting {self.subscriber_id}")

        # Load own position from Event store
        await self.load_position_on_start()

        # Start the polling loop
        self.loop.create_task(self.poll())

    async def poll(self) -> None:
        """
        Polling loop for processing messages.

        This method continuously polls for new messages and processes them by calling the `tick` method.
        It sleeps for the specified `tick_interval` between each tick.

        Returns:
            None
        """
        await self.tick()

        if self.keep_going and not self.engine.shutting_down:
            await asyncio.sleep(self.tick_interval)
            self.loop.create_task(self.poll())

    async def tick(self):
        """
        This method retrieves the next batch of messages to process and calls the `process_batch` method
        to handle each message. It also updates the read position after processing each message.

        Returns:
            None
        """
        messages = await self.get_next_batch_of_messages()
        if messages:
            await self.process_batch(messages)

    async def shutdown(self):
        """
        Shutdown the subscription.

        This method signals the subscription to stop polling and updates the current position to the store.
        It also logs a message indicating the shutdown of the subscription.

        Returns:
            None
        """
        self.keep_going = False  # Signal to stop polling
        await self.update_current_position_to_store()
        logger.info(f"Shutting down subscription {self.subscriber_id}")

    async def fetch_last_position(self):
        """
        Fetch the last read position from the store.

        Returns:
            int: The last read position from the store.
        """
        message = self.store._read_last_message(self.subscriber_stream_name)
        if message:
            return message["data"]["position"]

        return -1

    async def load_position_on_start(self) -> None:
        """
        Load the last position from the store when starting.

        This method retrieves the last read position from the event store and updates the current position
        of the subscription. If there is no previous position, it logs a message indicating that the
        subscription will start at the beginning of the stream.

        Returns:
            None
        """
        last_position = await self.fetch_last_position()
        if last_position > -1:
            self.current_position = last_position
            logger.debug(f"Loaded position {self.current_position} from last message")
        else:
            logger.debug(
                "No previous messages - Starting at the beginning of the stream"
            )

    async def update_current_position_to_store(self) -> int:
        """Update the current position to the store, only if out of sync.

        This method updates the current position of the subscription to the event store, but only if the
        current position is greater than the last written position.

        Returns:
            int: The last written position.
        """
        last_written_position = await self.fetch_last_position()
        if last_written_position < self.current_position:
            self.write_position(self.current_position)

        return last_written_position

    async def update_read_position(self, position) -> int:
        """
        Update the current read position.

        If at or beyond the configured interval, write position to the store.

        Args:
            position (int): The new read position.

        Returns:
            int: The updated read position.
        """
        self.current_position = position
        self.messages_since_last_position_write += 1

        if self.messages_since_last_position_write >= self.position_update_interval:
            self.write_position(position)

        return self.current_position

    def write_position(self, position: int) -> int:
        """
        Write the position to the store.

        This method writes the current read position to the event store. It updates the read position
        of the subscriber and resets the counter for messages since the last position write.

        Args:
            position (int): The read position to be written.

        Returns:
            int: The position that was written.
        """
        logger.debug(f"Updating Read Position of {self.subscriber_id} to {position}")

        self.messages_since_last_position_write = 0  # Reset counter

        return self.store._write(
            self.subscriber_stream_name,
            "Read",
            {"position": position},
            metadata={
                "kind": MessageType.READ_POSITION.value,
                "origin_stream_name": self.stream_name,
            },
        )

    def filter_on_origin(self, messages: List[Message]) -> List[Message]:
        """
        Filter messages based on the origin stream name.

        Args:
            messages (List[Message]): The list of messages to filter.

        Returns:
            List[Message]: The filtered list of messages.
        """
        if not self.origin_stream_name:
            return messages

        filtered_messages = []

        for message in messages:
            origin_stream = message.metadata and self.store.category(
                message.metadata.origin_stream_name
            )

            if self.origin_stream_name == origin_stream:
                filtered_messages.append(message)

        logger.debug(f"Filtered {len(filtered_messages)} out of {len(messages)}")
        return filtered_messages

    async def get_next_batch_of_messages(self):
        """
        Get the next batch of messages to process.

        This method reads messages from the event store starting from the current position + 1.
        It retrieves a specified number of messages per tick and applies filtering based on the origin stream name.

        Returns:
            List[Message]: The next batch of messages to process.
        """
        messages = self.store.read(
            self.stream_name,
            position=self.current_position + 1,
            no_of_messages=self.messages_per_tick,
        )  # FIXME Implement filtering

        return self.filter_on_origin(messages)

    async def process_batch(self, messages):
        """
        Process a batch of messages.

        This method takes a batch of messages and processes each message by calling the `handle_message` method
        of the engine. It also updates the read position after processing each message. If an exception occurs
        during message processing, it logs the error using the `log_error` method.

        Args:
            messages (List[Message]): The batch of messages to process.

        Returns:
            int: The number of messages processed.
        """
        logging.debug(f"Processing {len(messages)} messages...")
        for message in messages:
            logging.info(f"{message.type}-{message.id} : {message.to_dict()}")
            await self.engine.handle_message(self.handler, message)
            await self.update_read_position(message.global_position)

        return len(messages)



from __future__ import annotations

import asyncio
import logging
import signal
import traceback
from typing import Type, Union

from protean.core.command_handler import BaseCommandHandler
from protean.core.event_handler import BaseEventHandler
from protean.globals import g
from protean.utils.mixins import Message

from .subscription import Subscription

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s,%(msecs)d %(levelname)s: %(message)s",
    datefmt="%H:%M:%S",
)

logger = logging.getLogger(__name__)


class Engine:
    """
    The Engine class represents the Protean Engine that handles message processing and subscription management.
    """

    def __init__(self, domain, test_mode: bool = False, debug: bool = False) -> None:
        """
        Initialize the Engine.

        Modes:
        - Test Mode: If set to True, the engine will run in test mode and will exit after all tasks are completed.
        - Debug Mode: If set to True, the engine will run in debug mode and will log additional information.

        Args:
            domain (Domain): The domain object associated with the engine.
            test_mode (bool, optional): Flag to indicate if the engine is running in test mode. Defaults to False.
            debug (bool, optional): Flag to indicate if debug mode is enabled. Defaults to False.
        """
        self.domain = domain
        self.test_mode = (
            test_mode  # Flag to indicate if the engine is running in test mode
        )
        self.debug = debug  # Flag to indicate if debug mode is enabled
        self.exit_code = 0
        self.shutting_down = False  # Flag to indicate the engine is shutting down

        self.loop = asyncio.get_event_loop()

        # FIXME Gather all handlers
        self._subscriptions = {}
        for handler_name, record in self.domain.registry.event_handlers.items():
            # Create a subscription for each event handler
            self._subscriptions[handler_name] = Subscription(
                self,
                handler_name,
                record.cls.meta_.stream_name
                or record.cls.meta_.part_of.meta_.stream_name,
                record.cls,
                origin_stream_name=record.cls.meta_.source_stream,
            )

        for handler_name, record in self.domain.registry.command_handlers.items():
            # Create a subscription for each command handler
            self._subscriptions[handler_name] = Subscription(
                self,
                handler_name,
                f"{record.cls.meta_.part_of.meta_.stream_name}:command",
                record.cls,
            )

    async def handle_message(
        self,
        handler_cls: Type[Union[BaseCommandHandler, BaseEventHandler]],
        message: Message,
    ) -> None:
        """
        Handle a message by invoking the appropriate handler class.

        Args:
            handler_cls (Type[Union[BaseCommandHandler, BaseEventHandler]]): The handler class to invoke.
            message (Message): The message to be handled.

        Returns:
            None

        Raises:
            Exception: If an error occurs while handling the message.

        """
        if self.shutting_down:
            return  # Skip handling if shutdown is in progress

        with self.domain.domain_context():
            # Set context from current message, so that further processes
            #   carry the metadata forward.
            g.message_in_context = message

            try:
                handler_cls._handle(message)

                logger.info(
                    f"{handler_cls.__name__} processed {message.type}-{message.id} successfully."
                )
            except Exception as exc:  # Includes handling `ConfigurationError`
                logger.error(
                    f"Error handling message {message.stream_name}-{message.id} "
                    f"in {handler_cls.__name__}"
                )
                logger.error(f"{str(exc)}")
                handler_cls.handle_error(exc, message)

                await self.shutdown(exit_code=1)
                return

            # Reset message context
            g.pop("message_in_context")

    async def shutdown(self, signal=None, exit_code=0):
        """
        Cleanup tasks tied to the service's shutdown.

        Args:
            signal (Optional[signal]): The exit signal received. Defaults to None.
            exit_code (int): The exit code to be stored. Defaults to 0.
        """
        self.shutting_down = True  # Set shutdown flag

        try:
            if signal:
                logger.info(f"Received exit signal {signal.name}...")

            # Store the exit code
            self.exit_code = exit_code

            tasks = [t for t in asyncio.all_tasks() if t is not asyncio.current_task()]

            # Shutdown subscriptions
            subscription_shutdown_tasks = [
                subscription.shutdown()
                for _, subscription in self._subscriptions.items()
            ]

            # Cancel outstanding tasks
            [task.cancel() for task in tasks]
            logger.info(f"Cancelling {len(tasks)} outstanding tasks")
            await asyncio.gather(*tasks, return_exceptions=True)

            # Wait for subscriptions to shut down
            await asyncio.gather(*subscription_shutdown_tasks, return_exceptions=True)
            logger.info("All subscriptions have been shut down.")
        finally:
            if self.loop.is_running():
                self.loop.stop()

    def run(self):
        """
        Start the Protean Engine and run the subscriptions.
        """
        logger.info("Starting Protean Engine...")
        # Handle Signals
        signals = (signal.SIGHUP, signal.SIGTERM, signal.SIGINT)
        for s in signals:
            self.loop.add_signal_handler(
                s, lambda s=s: asyncio.create_task(self.shutdown(signal=s))
            )

        # Handle Exceptions
        def handle_exception(loop, context):
            # context["message"] will always be there; but context["exception"] may not
            msg = context.get("exception", context["message"])

            # Print the stack trace
            traceback.print_stack(context.get("exception"))

            logger.error(f"Caught exception: {msg}")
            logger.info("Shutting down...")
            if loop.is_running():
                asyncio.create_task(self.shutdown(exit_code=1))

        self.loop.set_exception_handler(handle_exception)

        if len(self._subscriptions) == 0:
            logger.info("No subscriptions to start. Exiting...")

        # Start consumption, one per subscription
        try:
            tasks = [
                self.loop.create_task(subscription.start())
                for _, subscription in self._subscriptions.items()
            ]

            if self.test_mode:
                # If in test mode, run until all tasks complete
                self.loop.run_until_complete(asyncio.gather(*tasks))
                # Then immediately call and await the shutdown directly
                self.loop.run_until_complete(self.shutdown())
            else:
                self.loop.run_forever()
        finally:
            self.loop.close()
            logger.debug("Successfully shutdown Protean Engine.")



from __future__ import annotations

import logging
from functools import lru_cache
from typing import TYPE_CHECKING

from protean.container import Element, OptionsMixin
from protean.core.aggregate import BaseAggregate
from protean.core.unit_of_work import UnitOfWork
from protean.exceptions import IncorrectUsageError, NotSupportedError
from protean.fields import HasMany, HasOne
from protean.globals import current_uow
from protean.port.dao import BaseDAO
from protean.port.provider import BaseProvider
from protean.reflection import association_fields, has_association_fields
from protean.utils import (
    Database,
    DomainObjects,
    derive_element_class,
    fully_qualified_name,
)

if TYPE_CHECKING:
    from protean.domain import Domain

logger = logging.getLogger(__name__)


class BaseRepository(Element, OptionsMixin):
    """This is the baseclass for concrete Repository implementations.

    The three methods in this baseclass to `add`, `get` or `all` entities are sufficient in most cases
    to handle application requirements. They have built-in support for handling child relationships and
    honor Unit of Work constructs. While they can be overridden, it is generally suggested to call the
    parent method first before writing custom code.

    Repositories are strictly meant to be used in conjunction with Aggregate elements. It is always prudent to deal
    with persistence at the transaction boundary, which is at an Aggregate's level.
    """

    element_type = DomainObjects.REPOSITORY

    @classmethod
    def _default_options(cls):
        return [("part_of", None), ("database", "ALL")]

    def __new__(cls, *args, **kwargs):
        # Prevent instantiation of `BaseRepository itself`
        if cls is BaseRepository:
            raise NotSupportedError("BaseRepository cannot be instantiated")
        return super().__new__(cls)

    def __init__(self, domain: Domain, provider: BaseProvider) -> None:
        self._domain = domain
        self._provider = provider

    @property
    @lru_cache()
    def _model(self):
        """Retrieve Model class connected to Entity"""
        # If a model was associated with the aggregate record, give it a higher priority
        #   and do not bake a new model class from aggregate/entity attributes
        custom_model_cls = None
        if fully_qualified_name(self.meta_.part_of) in self._domain._models:
            custom_model_cls = self._domain._models[
                fully_qualified_name(self.meta_.part_of)
            ]

        # FIXME This is the provide support for activating database specific models
        #   This needs to be enhanced to allow Protean to hold multiple models per Aggregate/Entity
        #   per database.
        #
        #   If no database is specified, model can be used for all databases
        if custom_model_cls and (
            custom_model_cls.meta_.database is None
            or custom_model_cls.meta_.database == self._provider.conn_info["database"]
        ):
            # Get the decorated model class.
            #   This is a no-op if the provider decides that the model is fully-baked
            model_cls = self._provider.decorate_model_class(
                self.meta_.part_of, custom_model_cls
            )
        else:
            # No model was associated with the aggregate/entity explicitly.
            #   So ask the Provider to bake a new model, initialized properly for this aggregate
            #   and return it
            model_cls = self._provider.construct_model_class(self.meta_.part_of)

        return model_cls

    @property
    @lru_cache()
    def _dao(self) -> BaseDAO:
        """Retrieve a DAO registered for the Aggregate with a live connection"""
        # Fixate on Model class at the domain level because an explicit model may have been registered
        return self._provider.get_dao(self.meta_.part_of, self._model)

    def add(self, aggregate: BaseAggregate) -> BaseAggregate:  # noqa: C901
        """This method helps persist or update aggregates into the persistence store.

        Returns the persisted aggregate.

        Protean adopts a collection-oriented design pattern to handle persistence. What this means is that
        the Repository interface does not hint in any way that there is an underlying persistence mechanism,
        avoiding any notion of saving or persisting data in the design layer. The task of syncing the data
        back into the persistence store is handled automatically.

        To be specific, a Repository mimics a `set` collection. Whatever the implementation, the repository
        will not allow instances of the same object to be added twice. Also, when retrieving objects from
        a Repository and modifying them, you don’t need to “re-save” them to the Repository.

        If there is a :ref:`Unit of Work <unit-of-work>` in progress, then the changes are performed on the
        UoW's active session. They are committed whenever the entire UoW is committed. If there is no
        transaction in progress, changes are committed immediately to the persistence store. This mechanism
        is part of the DAO's design, and is automatically used wherever one tries to persist data.
        """
        # `add` is typically invoked in handler methods in Command Handlers and Event Handlers, which are
        #   enclosed in a UoW automatically. Therefore, if there is a UoW in progress, we can assume
        #   that it is the active session. If not, we will start a new UoW and commit it after the operation
        #   is complete.
        own_current_uow = None
        if not (current_uow and current_uow.in_progress):
            own_current_uow = UnitOfWork()
            own_current_uow.start()

        # If there are HasMany/HasOne fields in the aggregate, sync child objects added/removed,
        if has_association_fields(aggregate):
            self._sync_children(aggregate)

        # Persist only if the aggregate object is new, or it has changed since last persistence
        if (not aggregate.state_.is_persisted) or (
            aggregate.state_.is_persisted and aggregate.state_.is_changed
        ):
            self._dao.save(aggregate)

        # If we started a UnitOfWork, commit it now
        if own_current_uow:
            own_current_uow.commit()

        return aggregate

    def _sync_children(self, entity):
        """Recursively sync child entities to the persistence store"""
        # If there are HasMany fields in the aggregate, sync child objects added/removed,
        #   but not yet persisted to the database.
        #
        # The details of in-transit child objects are maintained as part of the `has_many_field` itself
        #   in a variable called `_temp_cache`
        for field_name, field in association_fields(entity).items():
            ### RECURSIVE SYNC ###
            # Start at the innermost child and work our way up
            if has_association_fields(field.to_cls):
                if isinstance(field, HasMany):
                    for item in getattr(entity, field_name):
                        self._sync_children(item)
                elif isinstance(field, HasOne):
                    if getattr(entity, field_name):
                        self._sync_children(getattr(entity, field_name))
            ### RECURSIVE SYNC ###

            if isinstance(field, HasMany):
                # First, handle direct updates to underlying child objects
                #   These are ones whose attributes have been changed directly
                #   instead of being routed via `add`/`remove`
                for item in getattr(entity, field_name):
                    if item.state_.is_changed:
                        # If the item was changed directly AND added via `add`, then
                        #   we give preference to the object in the cache
                        if item not in entity._temp_cache[field_name]["updated"]:
                            self._domain.repository_for(field.to_cls)._dao.save(item)

                for _, item in entity._temp_cache[field_name]["removed"].items():
                    self._domain.repository_for(field.to_cls)._dao.delete(item)
                entity._temp_cache[field_name][
                    "removed"
                ] = {}  # Empty contents of `removed` cache

                for _, item in entity._temp_cache[field_name]["updated"].items():
                    self._domain.repository_for(field.to_cls)._dao.save(item)
                entity._temp_cache[field_name][
                    "updated"
                ] = {}  # Empty contents of `updated` cache

                for _, item in entity._temp_cache[field_name]["added"].items():
                    item.state_.mark_new()
                    self._domain.repository_for(field.to_cls)._dao.save(item)
                entity._temp_cache[field_name][
                    "added"
                ] = {}  # Empty contents of `added` cache

            if isinstance(field, HasOne):
                # First, handle direct updates to underlying child objects
                #   These are ones whose attributes have been changed directly
                #   instead of being routed via `add`/`remove`
                item = getattr(entity, field_name)
                to_cls_repo = self._domain.repository_for(field.to_cls)
                if item is not None and item.state_.is_changed:
                    to_cls_repo._dao.save(item)
                # Or a new instance has been assigned
                elif entity._temp_cache[field_name]["change"]:
                    if entity._temp_cache[field_name]["change"] == "ADDED":
                        to_cls_repo._dao.save(item)
                    elif entity._temp_cache[field_name]["change"] == "UPDATED":
                        if entity._temp_cache[field_name]["old_value"] is not None:
                            # The object was replaced, so delete the old record
                            to_cls_repo._dao.delete(
                                entity._temp_cache[field_name]["old_value"]
                            )
                        else:
                            # The same object was updated
                            # FIXME This should have been automatic with `is_changed` flag in `state_`
                            item.state_.mark_changed()

                        to_cls_repo._dao.save(item)
                    elif entity._temp_cache[field_name]["change"] == "DELETED":
                        to_cls_repo._dao.delete(
                            entity._temp_cache[field_name]["old_value"]
                        )

                    # Reset temporary fields after processing
                    entity._temp_cache[field_name]["change"] = None
                    entity._temp_cache[field_name]["old_value"] = None

    def get(self, identifier) -> BaseAggregate:
        """This is a utility method to fetch data from the persistence store by its key identifier. All child objects,
        including enclosed entities, are returned as part of this call.

        Returns the fetched aggregate.

        All other data filtering capabilities can be implemented by using the underlying DAO's
        :meth:`BaseDAO.filter` method.

        Filter methods are typically implemented as domain-contextual queries, like `find_adults()`,
        `find_residents_of_area(zipcode)`, etc. It is also possible to make use of more complicated,
        domain-friendly design patterns like the `Specification` pattern.
        """
        return self._dao.get(identifier)


def repository_factory(element_cls, **opts):
    element_cls = derive_element_class(element_cls, BaseRepository, **opts)

    if not element_cls.meta_.part_of:
        raise IncorrectUsageError(
            {
                "_entity": [
                    f"Repository `{element_cls.__name__}` should be associated with an Aggregate"
                ]
            }
        )

    # FIXME Uncomment
    # if not issubclass(element_cls.meta_.part_of, BaseAggregate):
    #     raise IncorrectUsageError(
    #         {"_entity": [f"Repository `{element_cls.__name__}` can only be associated with an Aggregate"]}
    #     )

    # Ensure the value of `database` is among known databases
    if element_cls.meta_.database != "ALL" and element_cls.meta_.database not in [
        database.value for database in Database
    ]:
        raise IncorrectUsageError(
            {
                "_entity": [
                    f"Repository `{element_cls.__name__}` should be associated with a valid Database"
                ]
            }
        )

    return element_cls



import logging

from protean.container import BaseContainer, OptionsMixin
from protean.exceptions import IncorrectUsageError, NotSupportedError
from protean.fields import Field
from protean.reflection import _ID_FIELD_NAME, declared_fields
from protean.utils import DomainObjects, derive_element_class

logger = logging.getLogger(__name__)


class BaseEvent(BaseContainer, OptionsMixin):  # FIXME Remove OptionsMixin
    """Base Event class that all Events should inherit from.

    Core functionality associated with Events, like timestamping, are specified
    as part of the base Event class.
    """

    element_type = DomainObjects.EVENT

    def __new__(cls, *args, **kwargs):
        if cls is BaseEvent:
            raise NotSupportedError("BaseEvent cannot be instantiated")
        return super().__new__(cls)

    def __init_subclass__(subclass) -> None:
        super().__init_subclass__()

        if not subclass.meta_.abstract:
            subclass.__track_id_field()

    def __setattr__(self, name, value):
        if not hasattr(self, "_initialized") or not self._initialized:
            return super().__setattr__(name, value)
        else:
            raise IncorrectUsageError(
                {
                    "_event": [
                        "Event Objects are immutable and cannot be modified once created"
                    ]
                }
            )

    @classmethod
    def _default_options(cls):
        return [
            ("abstract", False),
            ("part_of", None),
            ("stream_name", None),
            ("aggregate_cluster", None),
        ]

    @classmethod
    def __track_id_field(subclass):
        """Check if an identifier field has been associated with the command.

        When an identifier is provided, its value is used to construct
        unique stream name."""
        try:
            id_field = next(
                field
                for _, field in declared_fields(subclass).items()
                if isinstance(field, (Field)) and field.identifier
            )

            setattr(subclass, _ID_FIELD_NAME, id_field.field_name)

        except StopIteration:
            # No Identity fields declared
            pass


def domain_event_factory(element_cls, **kwargs):
    element_cls = derive_element_class(element_cls, BaseEvent, **kwargs)

    if (
        not (element_cls.meta_.part_of or element_cls.meta_.stream_name)
        and not element_cls.meta_.abstract
    ):
        raise IncorrectUsageError(
            {
                "_event": [
                    f"Event `{element_cls.__name__}` needs to be associated with an aggregate or a stream"
                ]
            }
        )

    return element_cls



"""Utility module for Protean

Definitions/declaractions in this module should be independent of other modules,
to the maximum extent possible.
"""

import functools
import importlib
import logging
from datetime import UTC, datetime
from enum import Enum
from typing import Any, Callable
from uuid import uuid4

from protean.exceptions import ConfigurationError
from protean.globals import current_domain

logger = logging.getLogger(__name__)


class IdentityStrategy(Enum):
    UUID = "uuid"
    FUNCTION = "function"


class IdentityType(Enum):
    INTEGER = "integer"
    STRING = "string"
    UUID = "uuid"


class EventProcessing(Enum):
    SYNC = "sync"
    ASYNC = "async"


class CommandProcessing(Enum):
    SYNC = "sync"
    ASYNC = "async"


class Database(Enum):
    elasticsearch = "elasticsearch"
    memory = "memory"
    postgresql = "postgresql"
    sqlite = "sqlite"


class Cache(Enum):
    memory = "memory"
    redis = "redis"


class TypeMatcher:
    """Allow assertion on object type.

    Ex. mocked_object.assert_called_once_with(TypeMatcher(TargetCls))
    """

    def __init__(self, expected_type):
        self.expected_type = expected_type

    def __eq__(self, other):
        return isinstance(other, self.expected_type)


def utcnow_func():
    """Return the current time in UTC with timezone information"""
    return datetime.now(UTC)


def get_version():
    return importlib.metadata.version("protean")


def import_from_full_path(domain, path):
    spec = importlib.util.spec_from_file_location(domain, path)
    mod = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(mod)

    return getattr(mod, domain)


def fully_qualified_name(cls):
    """Return Fully Qualified name along with module"""
    return ".".join([cls.__module__, cls.__name__])


fqn = fully_qualified_name


def singleton(cls):
    """Make a class a Singleton class (only one instance)"""

    @functools.wraps(cls)
    def wrapper_singleton(*args, **kwargs):
        if not wrapper_singleton.instance:
            wrapper_singleton.instance = cls(*args, **kwargs)
        return wrapper_singleton.instance

    wrapper_singleton.instance = None
    return wrapper_singleton


def convert_str_values_to_list(value):
    if not value:
        return []
    elif isinstance(value, str):
        return [value]
    else:
        return list(value)


class DomainObjects(Enum):
    AGGREGATE = "AGGREGATE"
    APPLICATION_SERVICE = "APPLICATION_SERVICE"
    COMMAND = "COMMAND"
    COMMAND_HANDLER = "COMMAND_HANDLER"
    EVENT = "EVENT"
    EVENT_HANDLER = "EVENT_HANDLER"
    EVENT_SOURCED_AGGREGATE = "EVENT_SOURCED_AGGREGATE"
    EVENT_SOURCED_REPOSITORY = "EVENT_SOURCED_REPOSITORY"
    DOMAIN_SERVICE = "DOMAIN_SERVICE"
    EMAIL = "EMAIL"
    ENTITY = "ENTITY"
    MODEL = "MODEL"
    REPOSITORY = "REPOSITORY"
    SERIALIZER = "SERIALIZER"
    SUBSCRIBER = "SUBSCRIBER"
    VALUE_OBJECT = "VALUE_OBJECT"
    VIEW = "VIEW"


def derive_element_class(element_cls, base_cls, **opts):
    from protean.container import Options

    if not issubclass(element_cls, base_cls):
        try:
            new_dict = element_cls.__dict__.copy()
            new_dict.pop("__dict__", None)  # Remove __dict__ to prevent recursion

            new_dict["meta_"] = Options(opts)

            element_cls = type(element_cls.__name__, (base_cls,), new_dict)
        except BaseException as exc:
            logger.debug("Error during Element registration:", repr(exc))
            raise
    else:
        element_cls.meta_ = Options(opts)

    # Assign default options for remaining items
    element_cls._set_defaults()

    return element_cls


def generate_identity(
    identity_strategy: str = None,
    identity_function: Callable[[], Any] | None = None,
    identity_type: str = None,
):
    """Generate Unique Identifier, based on configured strategy and type.

    If an identity type is provided, it will override the domain's configuration.
    """
    id_value = None

    # Consider strategy defined in the Auto field. If not provided, fall back to the
    #   domain's configuration.
    id_strategy = identity_strategy or current_domain.config["identity_strategy"]

    # UUID Strategy
    if id_strategy == IdentityStrategy.UUID.value:
        id_type = identity_type or current_domain.config["identity_type"]

        if id_type == IdentityType.INTEGER.value:
            id_value = uuid4().int
        elif id_type == IdentityType.STRING.value:
            id_value = str(uuid4())
        elif id_type == IdentityType.UUID.value:
            id_value = uuid4()
        else:
            raise ConfigurationError(f"Unknown Identity Type {id_type}")

    # Function Strategy
    elif id_strategy == IdentityStrategy.FUNCTION.value:
        # Run the function configured as part of the Auto field. If not provided, fall back
        #   to the function defined at the domain level.
        id_function = identity_function or current_domain._identity_function

        id_value = id_function()

    else:
        raise ConfigurationError(f"Unknown Identity Strategy {id_strategy}")

    return id_value


__all__ = [
    "Cache",
    "CommandProcessing",
    "Database",
    "DomainObjects",
    "EventProcessing",
    "IdentityStrategy",
    "IdentityType",
    "TypeMatcher",
    "convert_str_values_to_list",
    "derive_element_class",
    "fetch_element_cls_from_registry",
    "fully_qualified_name",
    "generate_identity",
    "get_version",
    "import_from_full_path",
    "singleton",
    "utcnow_func",
]



from datetime import UTC, datetime
from typing import Any, Dict, List

from protean.core.aggregate import BaseAggregate
from protean.core.repository import BaseRepository
from protean.globals import current_domain
from protean.port.event_store import BaseEventStore
from protean.utils.mixins import MessageMetadata, MessageRecord


class MemoryMessage(BaseAggregate, MessageRecord):
    pass


class MemoryMessageRepository(BaseRepository):
    def is_category(self, stream_name: str) -> bool:
        if not stream_name:
            return False

        return "-" not in stream_name

    def stream_version(self, stream_name: str):
        repo = current_domain.repository_for(MemoryMessage)
        results = (
            repo._dao.query.filter(stream_name=stream_name).order_by("-position").all()
        )

        return results.first.position if results.items else -1

    def write(
        self,
        stream_name: str,
        message_type: str,
        data: Dict,
        metadata: Dict = None,
        expected_version: int = None,
    ) -> int:
        # Fetch stream version
        _stream_version = self.stream_version(stream_name)

        if expected_version is not None and expected_version != _stream_version:
            raise ValueError(
                f"Wrong expected version: {expected_version} "
                f"(Stream: {stream_name}, Stream Version: {_stream_version})"
            )

        next_position = _stream_version + 1

        self.add(
            MemoryMessage(
                stream_name=stream_name,
                position=next_position,
                type=message_type,
                data=data,
                metadata=MessageMetadata(**metadata) if metadata else None,
                time=datetime.now(UTC),
            )
        )

        return next_position

    def read(
        self,
        stream_name: str,
        sql: str = None,
        position: int = 0,
        no_of_messages: int = 1000,
    ):
        repo = current_domain.repository_for(MemoryMessage)
        q = (
            repo._dao.query.filter(position__gte=position)
            .order_by("position")
            .limit(no_of_messages)
        )

        if stream_name == "$all":
            pass  # Don't filter on stream name
        elif self.is_category(stream_name):
            q = q.filter(stream_name__contains=stream_name)
        else:
            q = q.filter(stream_name=stream_name)

        items = q.all().items
        return [item.to_dict() for item in items]


class MemoryEventStore(BaseEventStore):
    def __init__(self, domain, conn_info) -> None:
        super().__init__("Memory", domain, conn_info)

        self.domain = domain
        self.domain.register(MemoryMessage, provider="memory")
        self.domain.register(MemoryMessageRepository, part_of=MemoryMessage)

    def _write(
        self,
        stream_name: str,
        message_type: str,
        data: Dict,
        metadata: Dict = None,
        expected_version: int = None,
    ) -> int:
        repo = self.domain.repository_for(MemoryMessage)
        return repo.write(stream_name, message_type, data, metadata, expected_version)

    def _read(
        self,
        stream_name: str,
        sql: str = None,
        position: int = 0,
        no_of_messages: int = 1000,
    ) -> List[Dict[str, Any]]:
        repo = self.domain.repository_for(MemoryMessage)
        return repo.read(stream_name, sql, position, no_of_messages)

    def _read_last_message(self, stream_name) -> Dict[str, Any]:
        repo = self.domain.repository_for(MemoryMessage)

        messages = repo.read(stream_name)
        return messages[-1] if messages else None

    def _data_reset(self) -> None:
        """Flush all events.

        Useful for running tests with a clean slate.
        """
        repo = self.domain.repository_for(MemoryMessage)
        repo._dao._delete_all()



"""Module for defining basic Field types of Entity"""

import datetime
from uuid import UUID

import bleach
from dateutil.parser import parse as date_parser

from protean.exceptions import InvalidOperationError, ValidationError
from protean.fields import Field, validators
from protean.fields.embedded import ValueObject
from protean.globals import current_domain
from protean.utils import IdentityType


class String(Field):
    """Concrete field implementation for the string type.

    :param max_length: The maximum allowed length for the field.
    :param min_length: The minimum allowed length for the field.

    FIXME Should max_length be optional?
    """

    default_error_messages = {
        "invalid": '{value}" value must be a string.',
    }

    def __init__(self, max_length=255, min_length=None, sanitize=True, **kwargs):
        self.min_length = min_length
        self.max_length = max_length
        self.sanitize = sanitize
        self.default_validators = [
            validators.MinLengthValidator(self.min_length),
            validators.MaxLengthValidator(self.max_length),
        ]
        super().__init__(**kwargs)

    def _cast_to_type(self, value):
        """Convert the value to its string representation"""
        value = value if isinstance(value, str) else str(value)

        return bleach.clean(value) if self.sanitize else value

    def as_dict(self, value):
        """Return JSON-compatible value of self"""
        return value

    def __repr__(self):
        # Generate repr values specific to this field
        values = self._generic_param_values_for_repr()
        if self.max_length != 255:
            values.append(f"max_length={self.max_length}")
        if self.min_length:
            values.append(f"min_length={self.min_length}")
        if not self.sanitize:
            values.append("sanitize=False")

        return f"{self.__class__.__name__}(" + ", ".join(values) + ")"


class Text(Field):
    """Concrete field implementation for the text type."""

    default_error_messages = {
        "invalid": '{value}" value must be a string.',
    }

    def __init__(self, sanitize=True, **kwargs):
        self.sanitize = sanitize
        super().__init__(**kwargs)

    def _cast_to_type(self, value):
        """Convert the value to its string representation"""
        value = value if isinstance(value, str) else str(value)

        return bleach.clean(value) if self.sanitize else value

    def as_dict(self, value):
        """Return JSON-compatible value of self"""
        return value

    def __repr__(self):
        # Generate repr values specific to this field
        values = self._generic_param_values_for_repr()
        if not self.sanitize:
            values.append("sanitize=False")

        return f"{self.__class__.__name__}(" + ", ".join(values) + ")"


class Integer(Field):
    """Concrete field implementation for the Integer type.

    :param min_value: The minimum allowed value for the field.
    :param max_value: The maximum allowed value for the field.

    """

    default_error_messages = {
        "invalid": '"{value}" value must be an integer.',
    }

    def __init__(self, min_value=None, max_value=None, **kwargs):
        self.min_value = min_value
        self.max_value = max_value
        self.default_validators = [
            validators.MinValueValidator(self.min_value),
            validators.MaxValueValidator(self.max_value),
        ]
        super().__init__(**kwargs)

    def _cast_to_type(self, value):
        """Convert the value to an int and raise error on failures"""
        try:
            if isinstance(value, str):
                if value.strip() == "":
                    return None
            return int(value)
        except (ValueError, TypeError):
            self.fail("invalid", value=value)

    def as_dict(self, value):
        """Return JSON-compatible value of self"""
        return value

    def __repr__(self):
        # Generate repr values specific to this field
        values = self._generic_param_values_for_repr()
        if self.max_value:
            values.append(f"max_value={self.max_value}")
        if self.min_value or self.min_value == 0:
            values.append(f"min_value={self.min_value}")

        return f"{self.__class__.__name__}(" + ", ".join(values) + ")"


class Float(Field):
    """Concrete field implementation for the Floating type.

    :param min_value: The minimum allowed value for the field.
    :param max_value: The maximum allowed value for the field.

    """

    default_error_messages = {
        "invalid": '"{value}" value must be floating point number.',
    }

    def __init__(self, min_value=None, max_value=None, **kwargs):
        self.min_value = min_value
        self.max_value = max_value
        self.default_validators = [
            validators.MinValueValidator(self.min_value),
            validators.MaxValueValidator(self.max_value),
        ]
        super().__init__(**kwargs)

    def _cast_to_type(self, value):
        """Convert the value to a float and raise error on failures"""
        try:
            return float(value)
        except (ValueError, TypeError):
            self.fail("invalid", value=value)

    def as_dict(self, value):
        """Return JSON-compatible value of self"""
        return value

    def __repr__(self):
        # Generate repr values specific to this field
        values = self._generic_param_values_for_repr()
        if self.max_value:
            values.append(f"max_value={self.max_value}")
        if self.min_value or self.min_value == 0.0:
            values.append(f"min_value={self.min_value}")

        return f"{self.__class__.__name__}(" + ", ".join(values) + ")"


class Boolean(Field):
    """Concrete field implementation for the Boolean type."""

    default_error_messages = {
        "invalid": '"{value}" value must be either True or False.',
    }

    def _cast_to_type(self, value):
        """Convert the value to a boolean and raise error on failures"""
        if value in (True, False):
            return bool(value)
        if value in ("t", "True", "1"):
            return True
        if value in ("f", "False", "0"):
            return False
        self.fail("invalid", value=value)

    def as_dict(self, value):
        """Return JSON-compatible value of self"""
        return value


class List(Field):
    """
    A field that represents a list of values.

    :param content_type: The type of the items in the list.
    :type content_type: Field, optional
    :param pickled: Whether the list should be pickled when stored, defaults to False.
    :type pickled: bool, optional
    """

    default_error_messages = {
        "invalid": '"{value}" value must be of list type.',
        "invalid_content": "Invalid value {value}",
    }

    def __init__(self, content_type=String, pickled=False, **kwargs):
        if content_type not in [
            Boolean,
            Date,
            DateTime,
            Float,
            Identifier,
            Integer,
            String,
            Text,
            Dict,
        ] and not isinstance(content_type, ValueObject):
            raise ValidationError({"content_type": ["Content type not supported"]})
        self.content_type = content_type
        self.pickled = pickled

        super().__init__(**kwargs)

    def _cast_to_type(self, value):
        """Raise errors if the value is not a list, or
        the items in the list are not of the right data type.
        """
        if not isinstance(value, list):
            self.fail("invalid", value=value)

        # Try to cast value into the destination type.
        #   Throw error if the underlying type does not support value.
        new_value = []
        try:
            for item in value:
                if isinstance(self.content_type, ValueObject):
                    new_value.append(self.content_type._load(item))
                else:
                    new_value.append(self.content_type()._load(item))
        except ValidationError:
            self.fail("invalid_content", value=value)

        return new_value

    def as_dict(self, value):
        """Return JSON-compatible value of self"""
        new_value = []
        for item in value:
            if isinstance(self.content_type, ValueObject):
                new_value.append(self.content_type.as_dict(item))
            else:
                new_value.append(self.content_type().as_dict(item))
        return new_value


class Dict(Field):
    """
    A field that represents a dictionary.

    :param pickled: Whether to store the dictionary as a pickled object.
    :type pickled: bool, optional
    """

    default_error_messages = {
        "invalid": '"{value}" value must be of dict type.',
    }

    def __init__(self, pickled=False, **kwargs):
        self.pickled = pickled

        super().__init__(**kwargs)

    def _cast_to_type(self, value):
        """Raise error if the value is not a dict"""

        # Lists are allowed in JSON because Postgres supports them.
        #   All other databases treat these columns as BLOB/Pickled Type.
        if not isinstance(value, (dict, list)):
            self.fail("invalid", value=value)
        return value

    def as_dict(self, value):
        """Return JSON-compatible value of self"""
        # FIXME Convert value of objects that the dict holds?
        return value


class Auto(Field):
    """
    Auto Field represents an automatically generated field value.

    Values of Auto-fields are generated automatically and cannot be set explicitly.
    They cannot be marked as `required` for this reason - Protean does not accept
    values supplied for Auto fields.

    Args:
        increment (bool): Flag indicating whether the field value should be incremented automatically.
    """

    def __init__(
        self,
        increment=False,
        identity_strategy: str = None,
        identity_function: str = None,
        identity_type: str = None,
        **kwargs,
    ):
        self.increment = increment
        self.identity_strategy = identity_strategy
        self.identity_function = identity_function
        self.identity_type = identity_type

        super().__init__(**kwargs)

        # Force set required to false
        self.required = False

    def __set__(self, instance, value):
        """An Identifier Field once set cannot be reset or changed.
        We override the ``__set__`` method and prevent setting of new value if one was
        already set and is different from the new value
        """
        existing_value = getattr(instance, self.field_name)
        if existing_value is not None and value != existing_value:
            raise InvalidOperationError("Identifiers cannot be changed once set")

        value = self._load(value)
        instance.__dict__[self.field_name] = value

        if hasattr(instance, "state_"):
            # Mark Entity as Dirty
            instance.state_.mark_changed()

    def _cast_to_type(self, value):
        """Perform no validation for auto fields. Return the value as is"""
        return value

    def as_dict(self, value):
        """Return JSON-compatible value of self"""
        if not value:
            return None

        return value if isinstance(value, int) else str(value)

    def __repr__(self):
        # Generate repr values specific to this field
        values = self._generic_param_values_for_repr()
        if self.increment:
            values.append("increment=True")

        return f"{self.__class__.__name__}(" + ", ".join(values) + ")"


class Identifier(Field):
    """
    Represents an identifier field in a domain entity.

    An identifier field is used to uniquely identify an entity within a domain.
    It can have different types such as UUID, string, or integer, depending on the configuration.

    :param identity_type: The type of the identifier field. If not provided, it will be picked from the domain config.
    :type identity_type: str, optional
    :raises ValidationError: If the provided identity type is not supported.
    """

    def __init__(self, identity_type=None, **kwargs):
        # Validate the identity type
        if identity_type and identity_type not in [
            id_type.value for id_type in IdentityType
        ]:
            raise ValidationError({"identity_type": ["Identity type not supported"]})

        self.identity_type = identity_type
        super().__init__(**kwargs)

    def _cast_to_type(self, value):
        """Verify that value is either a UUID, a String or an Integer"""
        # A Boolean value is tested for specifically because `isinstance(value, int)` is `True` for Boolean values
        if not (isinstance(value, (UUID, str, int))) or isinstance(value, bool):
            self.fail("invalid", value=value)

        # Fixate on IdentityType if not done already
        #   This happens the first time an identifier field instance is used.
        #   We don't try to fix this in the constructor because the Domain may not be available at that time.
        if self.identity_type is None:
            self.identity_type = current_domain.config["identity_type"]

        # Ensure that the value is of the right type
        if self.identity_type == IdentityType.UUID.value:
            if not isinstance(value, UUID):
                try:
                    value = UUID(value)
                except (ValueError, AttributeError):
                    self.fail("invalid", value=value)
        elif self.identity_type == IdentityType.INTEGER.value:
            if not isinstance(value, int):
                try:
                    value = int(value)
                except ValueError:
                    self.fail("invalid", value=value)
        elif self.identity_type == IdentityType.STRING.value:
            if not isinstance(value, str):
                value = str(value)
        else:
            raise ValidationError({"identity_type": ["Identity type not supported"]})

        return value

    def __set__(self, instance, value):
        """An Identifier Field once set cannot be reset or changed if it is the identity of the object.

        We override the ``__set__`` method and prevent setting of new value if one was
        already set and is different from the new value
        """
        if self.identifier is True:
            existing_value = getattr(instance, self.field_name)
            if existing_value is not None and value != existing_value:
                raise InvalidOperationError("Identifiers cannot be changed once set")

        super().__set__(instance, value)

    def as_dict(self, value):
        """Return JSON-compatible value of self"""
        if isinstance(value, UUID):
            return str(value)
        return value


class Method(Field):
    """Helper field for custom methods associated with serializer fields"""

    def __init__(self, method_name, **kwargs):
        self.method_name = method_name
        super().__init__(**kwargs)

    def _cast_to_type(self, value):
        """Perform no validation for identifier fields. Return the value as is"""
        return value

    def as_dict(self, value):
        """Return JSON-compatible value of self"""
        return value


class Nested(Field):
    """Helper field for nested objects associated with serializer fields"""

    def __init__(self, schema_name, many=False, **kwargs):
        self.schema_name = schema_name
        self.many = many
        super().__init__(**kwargs)

    def _cast_to_type(self, value):
        """Perform no validation for identifier fields. Return the value as is"""
        return value

    def as_dict(self, value):
        """Return JSON-compatible value of self"""
        return value

    def __repr__(self):
        # Generate repr values specific to this field
        values = []
        if self.schema_name:
            values.append(f"'{self.schema_name}'")

        values.extend(self._generic_param_values_for_repr())

        return f"{self.__class__.__name__}(" + ", ".join(values) + ")"


class Date(Field):
    """Concrete field implementation for the Date type."""

    default_error_messages = {
        "invalid": '"{value}" has an invalid date format.',
        "datetime": "Expected a date but got a datetime {value}.",
    }

    def _cast_to_type(self, value):
        """Convert the value to a date and raise error on failures"""
        if isinstance(value, str) and not value:
            return None

        if isinstance(value, datetime.datetime):
            self.fail("datetime", value=value)

        if isinstance(value, datetime.date):
            return value

        try:
            value = date_parser(value)

            if not (
                value.hour == 0
                and value.minute == 0
                and value.second == 0
                and value.microsecond == 0
            ):
                self.fail("datetime", value=value)

            return value.date()
        except ValueError:
            self.fail("invalid", value=value)

    def as_dict(self, value):
        """Return JSON-compatible value of self"""
        return str(value) if value else None


class DateTime(Field):
    """Concrete field implementation for the Datetime/Timestamp type."""

    def _cast_to_type(self, value):
        """Convert the value to a datetime and raise error on failures"""
        if isinstance(value, str) and not value:
            return None

        if isinstance(value, datetime.datetime):
            return value

        if isinstance(value, datetime.date):
            value = datetime.datetime(value.year, value.month, value.day)
            return value
        try:
            value = date_parser(value)
            return value
        except ValueError:
            self.fail("invalid", value=value)

    def as_dict(self, value):
        """Return JSON-compatible value of self"""
        return str(value) if value else None



from . import validators
from .association import HasMany, HasOne, Reference
from .base import Field, FieldBase
from .basic import (
    Auto,
    Boolean,
    Date,
    DateTime,
    Dict,
    Float,
    Identifier,
    Integer,
    List,
    Method,
    Nested,
    String,
    Text,
)
from .embedded import ValueObject

__all__ = [
    "Auto",
    "Boolean",
    "Date",
    "DateTime",
    "Dict",
    "Field",
    "FieldBase",
    "Float",
    "HasMany",
    "HasOne",
    "Identifier",
    "Integer",
    "List",
    "Method",
    "Nested",
    "Reference",
    "String",
    "Text",
    "ValueObject",
    "validators",
]



import json
from typing import Optional, Union

import redis

from protean.core.view import BaseView
from protean.port.cache import BaseCache
from protean.reflection import id_field
from protean.utils.inflection import underscore


class RedisCache(BaseCache):
    def __init__(self, name, domain, conn_info: dict):
        """Initialize Cache with Connection/Adapter details"""

        # FIXME Update cache value to REDIS
        # In case of `RedisCache`, the `cache` value will always be `redis`.
        conn_info["cache"] = "redis"
        super().__init__(name, domain, conn_info)

        self.r = redis.Redis.from_url(conn_info["URI"])

    def ping(self):
        return self.r.ping()

    def get_connection(self):
        return self.r

    def add(self, view: BaseView, ttl: Optional[Union[int, float]] = None) -> None:
        """Add view record to cache

        KEY: View ID
        Value: View Data (derived from `to_dict()`)

        TTL is in seconds. If not specified explicitly in method call,
        it is picked up from Redis broker configuration. In the absence of
        configuration, it is set to 300 seconds.

        Args:
            view (BaseView): View Instance containing data
            ttl (int, float, optional): Timeout in seconds. Defaults to None.
        """
        identifier = getattr(view, id_field(view).field_name)
        key = f"{underscore(view.__class__.__name__)}:::{identifier}"

        ttl = ttl or self.conn_info.get("TTL") or 300

        self.r.psetex(key, int(ttl * 1000), json.dumps(view.to_dict()))

    def get(self, key):
        view_name = key.split(":::")[0]
        view_cls = self._views[view_name]

        value = self.r.get(key)
        return view_cls(json.loads(value)) if value else None

    def get_all(self, key_pattern, last_position=0, count=25):
        # FIXME Validate count
        view_name = key_pattern.split(":::")[0]
        view_cls = self._views[view_name]

        cursor, values = self.r.scan(
            cursor=last_position, match=key_pattern, count=count
        )
        return [view_cls(json.loads(self.r.get(value))) for value in values]

    def count(self, key_pattern, count=25):
        values = self.r.scan_iter(match=key_pattern, count=count)
        return len(list(values))

    def remove(self, view):
        identifier = getattr(view, id_field(view).field_name)
        key = f"{underscore(view.__class__.__name__)}:::{identifier}"
        self.r.delete(key)

    def remove_by_key(self, key):
        self.r.delete(key)

    def remove_by_key_pattern(self, key_pattern):
        values = self.r.scan_iter(match=key_pattern)
        self.r.delete(*values)

    def flush_all(self):
        self.r.flushall()

    def set_ttl(self, key, ttl):
        self.r.pexpire(key, ttl * 1000)

    def get_ttl(self, key):
        return self.r.pttl(key)



from __future__ import annotations

import copy
import inspect
import logging
from collections import defaultdict
from typing import Any, Type, Union

from protean.exceptions import (
    InvalidDataError,
    NotSupportedError,
    ValidationError,
)
from protean.fields import Auto, Field, FieldBase, ValueObject
from protean.utils import generate_identity

from .reflection import _FIELDS, _ID_FIELD_NAME, attributes, declared_fields, fields

logger = logging.getLogger(__name__)


class Element:
    """Base class for all Protean elements"""


class Options:
    """Metadata info for the Container.

    Common options:
    - ``abstract``: Indicates that this is an abstract entity (Ignores all other meta options)
    """

    def __init__(self, opts: Union[dict, Type] = None) -> None:
        self._opts = set()

        if opts:
            # FIXME Remove support passing a class as opts after revamping BaseSerializer
            #   The `inspect.isclass` check will not be necessary
            if isinstance(opts, (self.__class__)) or inspect.isclass(opts):
                attributes = inspect.getmembers(
                    opts, lambda a: not (inspect.isroutine(a))
                )
                for attr in attributes:
                    if not (
                        attr[0].startswith("__") and attr[0].endswith("__")
                    ) and attr[0] not in ["_opts"]:
                        setattr(self, attr[0], attr[1])

                self.abstract = getattr(opts, "abstract", None) or False
            elif isinstance(opts, dict):
                for opt_name, opt_value in opts.items():
                    setattr(self, opt_name, opt_value)

                self.abstract = opts.get("abstract", None) or False
        else:
            # Common Meta attributes
            self.abstract = getattr(opts, "abstract", None) or False

    def __setattr__(self, __name: str, __value: Any) -> None:
        # Ignore if `_opts` is being set
        if __name != "_opts":
            self._opts.add(__name)

        super().__setattr__(__name, __value)

    def __delattr__(self, __name: str) -> None:
        self._opts.discard(__name)

        super().__delattr__(__name)

    def __eq__(self, other) -> bool:
        """Equivalence check based only on data."""
        if type(other) is not type(self):
            return False

        return self.__dict__ == other.__dict__

    def __hash__(self) -> int:
        """Overrides the default implementation and bases hashing on values"""
        return hash(frozenset(self.__dict__.items()))

    def __add__(self, other: Options) -> None:
        new_options = copy.copy(self)
        for opt in other._opts:
            setattr(new_options, opt, getattr(other, opt))

        return new_options


class OptionsMixin:
    def __init_subclass__(subclass) -> None:
        """Setup Options metadata on elements

        Args:
            subclass (Protean Element): Subclass to initialize with metadata
        """
        if not hasattr(subclass, "meta_"):
            setattr(subclass, "meta_", Options())

        # Assign default options
        subclass._set_defaults()

        super().__init_subclass__()

    @classmethod
    def _set_defaults(cls):
        # Assign default options for remaining items
        #   with the help of `_default_options()` method defined in the Element's Root.
        #   Element Roots are `Event`, `Subscriber`, `Repository`, and so on.
        for key, default in cls._default_options():
            # FIXME Should the `None` check be replaced with a SENTINEL check?
            if not (hasattr(cls.meta_, key) and getattr(cls.meta_, key) is not None):
                setattr(cls.meta_, key, default)


class ContainerMeta(type):
    """
    This base metaclass processes the class declaration and
    constructs a meta object that can be used to introspect
    the concrete Container class later.

    It also sets up a `meta_` attribute on the concrete class
    to an instance of Meta, either the default of one that is
    defined in the concrete class.
    """

    def __new__(mcs, name, bases, attrs, **kwargs):
        """Initialize Container MetaClass and load attributes"""

        # Ensure initialization is only performed for subclasses of Container
        # (excluding Container class itself).
        parents = [b for b in bases if isinstance(b, ContainerMeta)]
        if not parents:
            return super().__new__(mcs, name, bases, attrs)

        # Gather fields in the order specified, starting with base classes
        fields_dict = {}

        # ... from base classes first
        for base in reversed(bases):
            if hasattr(base, _FIELDS):
                for field_name, field_obj in fields(base).items():
                    fields_dict[field_name] = field_obj

        # ... Apply own fields next
        for attr_name, attr_obj in attrs.items():
            if isinstance(attr_obj, FieldBase):
                fields_dict[attr_name] = attr_obj

        # Gather all non-field attributes
        dup_attrs = {
            attr_name: attr_obj
            for attr_name, attr_obj in attrs.items()
            if attr_name not in fields_dict
        }

        # Insert fields in the order in which they were specified
        #   When field names overlap, the last specified field wins
        dup_attrs.update(fields_dict)

        # Store fields in a special field for later reference
        dup_attrs[_FIELDS] = fields_dict

        return super().__new__(mcs, name, bases, dup_attrs, **kwargs)


class BaseContainer(metaclass=ContainerMeta):
    """The Base class for Protean-Compliant Data Containers.

    Provides helper methods to custom define attributes, and find attribute names
    during runtime.
    """

    def __new__(cls, *args, **kwargs):
        if cls is BaseContainer:
            raise NotSupportedError("BaseContainer cannot be instantiated")
        return super().__new__(cls)

    def __init__(self, *template, **kwargs):  # noqa: C901
        """
        Initialise the container.

        During initialization, set value on fields if validation passes.

        This initialization technique supports keyword arguments as well as dictionaries. You
            can even use a template for initial data.
        """
        self._initialized = False

        if self.meta_.abstract is True:
            raise NotSupportedError(
                f"{self.__class__.__name__} class has been marked abstract"
                f" and cannot be instantiated"
            )

        self.errors = defaultdict(list)

        # Load the attributes based on the template
        loaded_fields = []
        for dictionary in template:
            if not isinstance(dictionary, dict):
                raise AssertionError(
                    f'Positional argument "{dictionary}" passed must be a dict.'
                    f"This argument serves as a template for loading common "
                    f"values.",
                )
            for field_name, val in dictionary.items():
                loaded_fields.append(field_name)
                setattr(self, field_name, val)

        # Now load against the keyword arguments
        for field_name, val in kwargs.items():
            try:
                setattr(self, field_name, val)
            except ValidationError as err:
                for field_name in err.messages:
                    self.errors[field_name].extend(err.messages[field_name])
            finally:
                loaded_fields.append(field_name)

        # Load Value Objects from associated fields
        #   This block will dynamically construct value objects from field values
        #   and associated the vo with the entity
        # If the value object was already provided, it will not be overridden.
        for field_name, field_obj in fields(self).items():
            if isinstance(field_obj, (ValueObject)) and not getattr(self, field_name):
                attrs = [
                    (embedded_field.field_name, embedded_field.attribute_name)
                    for embedded_field in field_obj.embedded_fields.values()
                ]
                values = {name: kwargs.get(attr) for name, attr in attrs}
                try:
                    value_object = field_obj.value_object_cls(**values)
                    # Set VO value only if the value object is not None/Empty
                    if value_object:
                        setattr(self, field_name, value_object)
                        loaded_fields.append(field_name)
                except ValidationError as err:
                    for sub_field_name in err.messages:
                        self.errors["{}_{}".format(field_name, sub_field_name)].extend(
                            err.messages[sub_field_name]
                        )

        # Load Identities
        for field_name, field_obj in declared_fields(self).items():
            if type(field_obj) is Auto and not field_obj.increment:
                if not getattr(self, field_obj.field_name, None):
                    setattr(
                        self,
                        field_obj.field_name,
                        generate_identity(
                            field_obj.identity_strategy,
                            field_obj.identity_function,
                            field_obj.identity_type,
                        ),
                    )
                loaded_fields.append(field_obj.field_name)

        # Now load the remaining fields with a None value, which will fail
        # for required fields
        for field_name in fields(self):
            if field_name not in loaded_fields:
                setattr(self, field_name, None)

        self.defaults()

        self._initialized = True

        # Raise any errors found during load
        if self.errors:
            logger.error(self.errors)
            raise ValidationError(self.errors)

    def defaults(self):
        """Placeholder method for defaults.
        To be overridden in concrete Containers, when an attribute's default depends on other attribute values.
        """

    def __eq__(self, other):
        """Equivalence check for containers is based only on data.

        Two container objects are considered equal if they have the same data.
        """
        if type(other) is not type(self):
            return False

        return self.to_dict() == other.to_dict()

    def __hash__(self):
        """Overrides the default implementation and bases hashing on values"""
        return hash(frozenset(self.to_dict().items()))

    def __repr__(self):
        """Friendly repr for Command"""
        return "<%s: %s>" % (self.__class__.__name__, self)

    def __str__(self):
        return "%s object (%s)" % (
            self.__class__.__name__,
            "{}".format(self.to_dict()),
        )

    def __bool__(self):
        """Return this object's truthiness to be `False`,
        if all its attributes evaluate to truthiness `False`
        """
        return any(bool(getattr(self, field_name, None)) for field_name in fields(self))

    def __setattr__(self, name, value):
        if (
            name in attributes(self)
            or name in fields(self)
            or name
            in [
                "errors",  # Errors in state transition
                "state_",  # Tracking dirty state of the entity
                "_temp_cache",  # Temporary cache (Assocations) for storing data befor persisting
                "_events",  # Temp placeholder for events raised by the entity
                "_initialized",  # Flag to indicate if the entity has been initialized
                "_root",  # Root entity in the hierarchy
                "_owner",  # Owner entity in the hierarchy
                "_disable_invariant_checks",  # Flag to disable invariant checks
            ]
            or name.startswith(("add_", "remove_", "get_one_from_", "filter_"))
        ):
            super().__setattr__(name, value)
        else:
            raise InvalidDataError({name: ["is invalid"]})

    def to_dict(self):
        """Return data as a dictionary"""
        return {
            field_name: field_obj.as_dict(getattr(self, field_name, None))
            for field_name, field_obj in fields(self).items()
        }

    def clone(self):
        """Deepclone the command"""
        return copy.deepcopy(self)

    @classmethod
    def _default_options(cls):
        # FIXME Raise exception
        # raise NotImplementedError
        return []


class EventedMixin:
    def __init__(self, *args, **kwargs) -> None:
        """Initialize an instance-level variable named `_events` to track events
        raised in the aggregate cluster.

        This method cannot have a super invocation, because we don't want it to
        invoke BaseContainer's `__init__` method. But there is a conflict regarding
        this between BaseAggregate and BaseEventSourcedAggregate. So this Mixin's
        functionality has been replicated temporarily in BaseAggregate class.

        Other mixins that are inherited by BaseEntity and BaseEventSourcedAggregate
        work with `__init_subclass__`, and do not have this issue.
        """
        super().__init__(*args, **kwargs)
        self._events = []

    def raise_(self, event) -> None:
        self._events.append(event)


class IdentityMixin:
    def __init_subclass__(subclass) -> None:
        super().__init_subclass__()

        # FIXME Is there a better way to check this?
        if subclass.__name__ not in [
            "BaseAggregate",
            "BaseEntity",
            "BaseEventSourcedAggregate",
        ]:
            subclass.__set_id_field()

    @classmethod
    def __set_id_field(new_class):
        """Lookup the id field for this entity and assign"""
        # FIXME What does it mean when there are no declared fields?
        #   Does it translate to an abstract entity?
        id_fields = [
            field
            for _, field in declared_fields(new_class).items()
            if isinstance(field, Field) and field.identifier
        ]

        if len(id_fields) > 1:
            raise NotSupportedError(
                {
                    "_entity": [
                        f"Multiple identifier fields found in entity {new_class.__name__}. "
                        f"Only one identifier field is allowed."
                    ]
                }
            )

        elif len(id_fields) == 1:
            # Remember the identity field
            setattr(new_class, _ID_FIELD_NAME, id_fields[0].field_name)

        else:
            # If no id field is declared then create one
            #   If entity is explicitly marked with `auto_add_id_field=False`,
            #   avoid creating an identifier field.
            if new_class.meta_.auto_add_id_field:
                new_class.__create_id_field()

    @classmethod
    def __create_id_field(new_class):
        """Create and return a default ID field that is Auto generated"""
        id_field = Auto(identifier=True)

        setattr(new_class, "id", id_field)

        # Set the name of the field on itself
        id_field.__set_name__(new_class, "id")

        # Set the name of the attribute on the class
        setattr(new_class, _ID_FIELD_NAME, id_field.field_name)

        # Add the attribute to _FIELDS for introspection
        field_objects = getattr(new_class, _FIELDS)
        field_objects["id"] = id_field
        setattr(new_class, _FIELDS, field_objects)



import functools
import inspect
import logging
import typing
from collections import defaultdict
from typing import Dict

from protean.container import BaseContainer, EventedMixin, IdentityMixin, OptionsMixin
from protean.core.event import BaseEvent
from protean.exceptions import IncorrectUsageError, NotSupportedError
from protean.fields import Integer
from protean.reflection import id_field
from protean.utils import (
    DomainObjects,
    derive_element_class,
    fully_qualified_name,
    inflection,
)

logger = logging.getLogger(__name__)


class BaseEventSourcedAggregate(
    OptionsMixin, IdentityMixin, EventedMixin, BaseContainer
):
    """Base Event Sourced Aggregate class that all EventSourced Aggregates should inherit from.

    The order of inheritance is important. We want BaseContainer to be initialised first followed by
    OptionsMixin (so that `meta_` is in place) before inheriting other mixins."""

    element_type = DomainObjects.EVENT_SOURCED_AGGREGATE

    # Track current version of Aggregate
    _version = Integer(default=-1)

    def __new__(cls, *args, **kwargs):
        if cls is BaseEventSourcedAggregate:
            raise NotSupportedError("BaseEventSourcedAggregate cannot be instantiated")
        return super().__new__(cls)

    @classmethod
    def _default_options(cls):
        return [
            ("auto_add_id_field", True),
            ("stream_name", inflection.underscore(cls.__name__)),
            ("aggregate_cluster", None),
        ]

    def __init_subclass__(subclass) -> None:
        super().__init_subclass__()

        # Associate a `_projections` map with subclasses.
        #   It needs to be initialized here because if it
        #   were initialized in __init__, the same collection object
        #   would be made available across all subclasses,
        #   defeating its purpose.
        setattr(subclass, "_projections", defaultdict(set))

        # Store associated events
        setattr(subclass, "_events_cls_map", {})

    def __eq__(self, other):
        """Equivalence check to be based only on Identity"""

        # FIXME Enhanced Equality Checks
        #   * Ensure IDs have values and both of them are not null
        #   * Ensure that the ID is of the right type
        #   * Ensure that Objects belong to the same `type`
        #   * Check Reference equality

        # FIXME Check if `==` and `in` operator work with __eq__

        if type(other) is type(self):
            self_id = getattr(self, id_field(self).field_name)
            other_id = getattr(other, id_field(other).field_name)

            return self_id == other_id

        return False

    def __hash__(self):
        """Overrides the default implementation and bases hashing on identity"""

        # FIXME Add Object Class Type to hash
        return hash(getattr(self, id_field(self).field_name))

    def _apply(self, event_dict: Dict) -> None:
        """Apply the event onto the aggregate by calling the appropriate projection.

        Args:
            event (BaseEvent): Event object to apply
        """
        # FIXME Handle case of missing projection
        for fn in self._projections[event_dict["type"]]:
            # Reconstruct Event object
            event_cls = self._events_cls_map[event_dict["type"]]
            event = event_cls(**event_dict["data"])

            # Call event handler method
            fn(self, event)


def apply(fn):
    """Decorator to mark methods in EventHandler classes."""

    if len(typing.get_type_hints(fn)) > 2:
        raise IncorrectUsageError(
            {
                "_entity": [
                    f"Apply method {fn.__name__} has incorrect number of arguments"
                ]
            }
        )

    try:
        _event_cls = next(
            iter(
                {
                    value
                    for value in typing.get_type_hints(fn).values()
                    if inspect.isclass(value) and issubclass(value, BaseEvent)
                }
            )
        )
    except StopIteration:
        raise IncorrectUsageError(
            {
                "_entity": [
                    f"Apply method `{fn.__name__}` should accept an argument annotated with the Event class"
                ]
            }
        )

    @functools.wraps(fn)
    def wrapper(*args):
        fn(*args)

    setattr(wrapper, "_event_cls", _event_cls)

    return wrapper


def event_sourced_aggregate_factory(element_cls, **opts):
    element_cls = derive_element_class(element_cls, BaseEventSourcedAggregate, **opts)

    # Iterate through methods marked as `@apply` and construct a projections map
    methods = inspect.getmembers(element_cls, predicate=inspect.isroutine)
    for method_name, method in methods:
        if not (
            method_name.startswith("__") and method_name.endswith("__")
        ) and hasattr(method, "_event_cls"):
            element_cls._projections[fully_qualified_name(method._event_cls)].add(
                method
            )
            element_cls._events_cls_map[fully_qualified_name(method._event_cls)] = (
                method._event_cls
            )

            # Associate Event with the aggregate class
            #
            # This can potentially cause a problem because an Event can only be associated
            #   with one aggregate class, but multiple event handlers can consume it.
            #   By resetting the event's aggregate class, its previous association is lost.
            #   We catch this problem during domain validation.
            #
            #   The domain validation should check for the same event class being present
            #   in `_events_cls_map` of multiple aggregate classes.
            if inspect.isclass(method._event_cls) and issubclass(
                method._event_cls, BaseEvent
            ):
                method._event_cls.meta_.part_of = element_cls

    return element_cls
